import itertools
import os
from os.path import exists
from sys import stderr
from typing import Dict, Generator, List, Optional, Callable, Set

from kgx.config import get_logger
from kgx.source import (
    GraphSource,
    Source,
    TsvSource,
    JsonSource,
    JsonlSource,
    ObographSource,
    TrapiSource,
    NeoSource,
    RdfSource,
    OwlSource,
    SssomSource,
)
from kgx.sink import (
    Sink,
    GraphSink,
    TsvSink,
    JsonSink,
    JsonlSink,
    NeoSink,
    RdfSink,
    NullSink,
)
from kgx.utils.kgx_utils import (
    apply_graph_operations,
    GraphEntityType,
    knowledge_provenance_properties,
)

SOURCE_MAP = {
    "tsv": TsvSource,
    "csv": TsvSource,
    "graph": GraphSource,
    "json": JsonSource,
    "jsonl": JsonlSource,
    "obojson": ObographSource,
    "obo-json": ObographSource,
    "trapi-json": TrapiSource,
    "neo4j": NeoSource,
    "nt": RdfSource,
    "owl": OwlSource,
    "sssom": SssomSource,
}

SINK_MAP = {
    "tsv": TsvSink,
    "csv": TsvSink,
    "graph": GraphSink,
    "json": JsonSink,
    "jsonl": JsonlSink,
    "neo4j": NeoSink,
    "nt": RdfSink,
    "null": NullSink,
}


log = get_logger()


class Transformer(object):
    """
    The Transformer class is responsible for transforming data from one
    form to another.

    Parameters
    ----------
    stream: bool
        Whether or not to stream
    infores_catalog: Optional[str]
        Optional dump of a TSV file of InfoRes CURIE to Knowledge Source mappings

    """

    def __init__(self, stream: bool = False, infores_catalog: Optional[str] = None):
        self.stream = stream
        self.node_filters = {}
        self.edge_filters = {}

        self.inspector: Optional[Callable[[GraphEntityType, List], None]] = None

        self.store = self.get_source("graph")
        self._seen_nodes = set()
        self._infores_catalog: Dict[str, str] = dict()

        if infores_catalog and exists(infores_catalog):
            with open(infores_catalog, "r") as irc:
                for entry in irc:
                    if len(entry):
                        entry = entry.strip()
                        if entry:
                            print("entry: " + entry, file=stderr)
                            source, infores = entry.split("\t")
                            self._infores_catalog[source] = infores

    def transform(
        self,
        input_args: Dict,
        output_args: Optional[Dict] = None,
        inspector: Optional[Callable[[GraphEntityType, List], None]] = None,
    ) -> None:
        """
        Transform an input source and write to an output sink.

        If ``output_args`` is not defined then the data is persisted to
        an in-memory graph.

        The 'inspector' argument is an optional Callable which the
        transformer.process() method applies to 'inspect' source records
        prior to writing them out to the Sink. The first (GraphEntityType)
        argument of the Callable tags the record as a NODE or an EDGE.
        The second argument given to the Callable is the current record
        itself. This Callable is strictly meant to be procedural and should
        *not* mutate the record.

        Parameters
        ----------
        input_args: Dict
            Arguments relevant to your input source
        output_args: Optional[Dict]
            Arguments relevant to your output sink (
        inspector: Optional[Callable[[GraphEntityType, List], None]]
            Optional Callable to 'inspect' source records during processing.
        """
        sources = []
        generators = []
        input_format = input_args["format"]
        prefix_map = input_args.pop("prefix_map", {})
        predicate_mappings = input_args.pop("predicate_mappings", {})
        node_property_predicates = input_args.pop("node_property_predicates", {})
        node_filters = input_args.pop("node_filters", {})
        edge_filters = input_args.pop("edge_filters", {})
        operations = input_args.pop("operations", [])

        # Optional process() data stream inspector
        self.inspector = inspector

        if input_format in {"neo4j", "graph"}:
            source = self.get_source(input_format)
            source.set_prefix_map(prefix_map)
            source.set_node_filters(node_filters)
            self.node_filters = source.node_filters
            self.edge_filters = source.edge_filters
            source.set_edge_filters(edge_filters)
            self.node_filters = source.node_filters
            self.edge_filters = source.edge_filters

            if "uri" in input_args:
                default_provenance = input_args["uri"]
            else:
                default_provenance = None

            g = source.parse(default_provenance=default_provenance, **input_args)

            sources.append(source)
            generators.append(g)
        else:
            filename = input_args.pop("filename", {})
            for f in filename:
                source = self.get_source(input_format)
                source.set_prefix_map(prefix_map)
                if isinstance(source, RdfSource):
                    source.set_predicate_mapping(predicate_mappings)
                    source.set_node_property_predicates(node_property_predicates)
                source.set_node_filters(node_filters)
                self.node_filters = source.node_filters
                self.edge_filters = source.edge_filters
                source.set_edge_filters(edge_filters)
                self.node_filters = source.node_filters
                self.edge_filters = source.edge_filters

                default_provenance = os.path.basename(f)

                g = source.parse(f, default_provenance=default_provenance, **input_args)

                sources.append(source)
                generators.append(g)

        source_generator = itertools.chain(*generators)

        if output_args:
            if self.stream:
                if output_args["format"] in {"tsv", "csv"}:
                    if "node_properties" not in output_args:
                        log.warning(
                            f"'node_properties' not defined for output while streaming. "
                            f"The exported {output_args['format']} will be limited to a subset of the columns."
                        )
                    if "edge_properties" not in output_args:
                        log.warning(
                            f"'edge_properties' not defined for output while streaming. "
                            f"The exported {output_args['format']} will be limited to a subset of the columns."
                        )
                sink = self.get_sink(**output_args)
                if "reverse_prefix_map" in output_args:
                    sink.set_reverse_prefix_map(output_args["reverse_prefix_map"])
                if isinstance(sink, RdfSink):
                    if "reverse_predicate_mapping" in output_args:
                        sink.set_reverse_predicate_mapping(
                            output_args["reverse_predicate_mapping"]
                        )
                    if "property_types" in output_args:
                        sink.set_property_types(output_args["property_types"])
                # stream from source to sink
                self.process(source_generator, sink)
                sink.finalize()
            else:
                # stream from source to intermediate
                intermediate_sink = GraphSink(self.store.graph)
                intermediate_sink.node_properties.update(self.store.node_properties)
                intermediate_sink.edge_properties.update(self.store.edge_properties)
                self.process(source_generator, intermediate_sink)
                for s in sources:
                    intermediate_sink.node_properties.update(s.node_properties)
                    intermediate_sink.edge_properties.update(s.edge_properties)
                apply_graph_operations(intermediate_sink.graph, operations)
                # stream from intermediate to output sink
                intermediate_source = self.get_source("graph")
                intermediate_source.node_properties.update(
                    intermediate_sink.node_properties
                )
                intermediate_source.edge_properties.update(
                    intermediate_sink.edge_properties
                )

                # Need to propagate knowledge source specifications here?
                ks_args = dict()
                for ksf in knowledge_provenance_properties:
                    if ksf in input_args:
                        ks_args[ksf] = input_args[ksf]

                # TODO: does this call also need the default_provenance named argument?
                intermediate_source_generator = intermediate_source.parse(
                    intermediate_sink.graph, **ks_args
                )

                if output_args["format"] in {"tsv", "csv"}:
                    if "node_properties" not in output_args:
                        output_args[
                            "node_properties"
                        ] = intermediate_source.node_properties
                    if "edge_properties" not in output_args:
                        output_args[
                            "edge_properties"
                        ] = intermediate_source.edge_properties
                    sink = self.get_sink(**output_args)
                    if "reverse_prefix_map" in output_args:
                        sink.set_reverse_prefix_map(output_args["reverse_prefix_map"])
                    if isinstance(sink, RdfSink):
                        if "reverse_predicate_mapping" in output_args:
                            sink.set_reverse_predicate_mapping(
                                output_args["reverse_predicate_mapping"]
                            )
                    if "property_types" in output_args:
                        sink.set_property_types(output_args["property_types"])
                else:
                    sink = self.get_sink(**output_args)
                    sink.node_properties.update(intermediate_source.node_properties)
                    sink.edge_properties.update(intermediate_source.edge_properties)

                self.process(intermediate_source_generator, sink)
                sink.finalize()
                self.store.node_properties.update(sink.node_properties)
                self.store.edge_properties.update(sink.edge_properties)
        else:
            # stream from source to intermediate
            sink = GraphSink(self.store.graph)
            self.process(source_generator, sink)
            sink.node_properties.update(self.store.node_properties)
            sink.edge_properties.update(self.store.edge_properties)
            for s in sources:
                sink.node_properties.update(s.node_properties)
                sink.edge_properties.update(s.edge_properties)
            sink.finalize()
            self.store.node_properties.update(sink.node_properties)
            self.store.edge_properties.update(sink.edge_properties)
            apply_graph_operations(sink.graph, operations)

        # Aggregate the InfoRes catalogs from  all sources
        for s in sources:
            for k, v in s.get_infores_catalog().items():
                self._infores_catalog[k] = v

    def get_infores_catalog(self):
        """
        Return catalog of Information Resource mappings
         aggregated from all Transformer associated sources
        """
        return self._infores_catalog

    def process(self, source: Generator, sink: Sink) -> None:
        """
        This method is responsible for reading from ``source``
        and writing to ``sink`` by calling the relevant methods
        based on the incoming data.

        .. note::
            The streamed data must not be mutated.

        Parameters
        ----------
        source: Generator
            A generator from a Source
        sink: kgx.sink.sink.Sink
            An instance of Sink

        """
        for rec in source:
            if rec:
                if len(rec) == 4:  # infer an edge record
                    write_edge = True
                    if "subject_category" in self.edge_filters:
                        if rec[0] in self._seen_nodes:
                            write_edge = True
                        else:
                            write_edge = False
                    if "object_category" in self.edge_filters:
                        if rec[1] in self._seen_nodes:
                            if "subject_category" in self.edge_filters:
                                if write_edge:
                                    write_edge = True
                            else:
                                write_edge = True
                        else:
                            write_edge = False
                    if write_edge:
                        if self.inspector:
                            self.inspector(GraphEntityType.EDGE, rec)
                        sink.write_edge(rec[-1])
                else:  # infer a node record
                    if "category" in self.node_filters:
                        self._seen_nodes.add(rec[0])
                    if self.inspector:
                        self.inspector(GraphEntityType.NODE, rec)
                    sink.write_node(rec[-1])

    # TODO: review whether or not the 'save()' method need to be 'knowledge_source' aware?
    def save(self, output_args: Dict) -> None:
        """
        Save data from the in-memory store to a desired sink.

        Parameters
        ----------
        output_args: Dict
            Arguments relevant to your output sink

        """
        if not self.store:
            raise Exception("self.store is empty.")
        source = self.store
        source.node_properties.update(self.store.node_properties)
        source.edge_properties.update(self.store.edge_properties)
        source_generator = source.parse(self.store.graph)
        if "node_properties" not in output_args:
            output_args["node_properties"] = source.node_properties
        if "edge_properties" not in output_args:
            output_args["edge_properties"] = source.edge_properties
        sink = self.get_sink(**output_args)
        sink.node_properties.update(source.node_properties)
        sink.edge_properties.update(source.edge_properties)
        if "reverse_prefix_map" in output_args:
            sink.set_reverse_prefix_map(output_args["reverse_prefix_map"])
        if isinstance(sink, RdfSink):
            if "reverse_predicate_mapping" in output_args:
                sink.set_reverse_predicate_mapping(
                    output_args["reverse_predicate_mapping"]
                )
            if "property_types" in output_args:
                sink.set_property_types(output_args["property_types"])
        self.process(source_generator, sink)
        sink.finalize()

    def get_source(self, format: str) -> Source:
        """
        Get an instance of Source that corresponds to a given format.

        Parameters
        ----------
        format: str
            The input store format

        Returns
        -------
        Source:
            An instance of kgx.source.Source

        """
        if format in SOURCE_MAP:
            s = SOURCE_MAP[format]
            return s()
        else:
            raise TypeError(f"{format} in an unrecognized format")

    def get_sink(self, **kwargs: Dict) -> Sink:
        """
        Get an instance of Sink that corresponds to a given format.

        Parameters
        ----------
        kwargs: Dict
            Arguments required for initializing an instance of Sink

        Returns
        -------
        Sink:
            An instance of kgx.sink.Sink

        """
        if kwargs["format"] in SINK_MAP:
            s = SINK_MAP[kwargs["format"]]
            return s(**kwargs)
        else:
            raise TypeError(f"{kwargs['format']} in an unrecognized format")



import re
import tarfile
from typing import Dict, Tuple, Any, Generator, Optional, List
import pandas as pd

from kgx.config import get_logger
from kgx.source.source import Source
from kgx.utils.kgx_utils import (
    generate_uuid,
    generate_edge_key,
    extension_types,
    archive_read_mode,
    sanitize_import,
    validate_edge,
    validate_node,
)

log = get_logger()


class TsvSource(Source):
    """
    TsvSource is responsible for reading data as records
    from a TSV/CSV.
    """

    def __init__(self):
        super().__init__()

    def set_prefix_map(self, m: Dict) -> None:
        """
        Add or override default prefix to IRI map.

        Parameters
        ----------
        m: Dict
            Prefix to IRI map

        """
        self.prefix_manager.set_prefix_map(m)

    def set_reverse_prefix_map(self, m: Dict) -> None:
        """
        Add or override default IRI to prefix map.

        Parameters
        ----------
        m: Dict
            IRI to prefix map

        """
        self.prefix_manager.set_reverse_prefix_map(m)

    def parse(
        self,
        filename: str,
        format: str,
        compression: Optional[str] = None,
        **kwargs: Any,
    ) -> Generator:
        """
        This method reads from a TSV/CSV and yields records.

        Parameters
        ----------
        filename: str
            The filename to parse
        format: str
            The format (``tsv``, ``csv``)
        compression: Optional[str]
            The compression type (``tar``, ``tar.gz``)
        kwargs: Any
            Any additional arguments

        Returns
        -------
        Generator
            A generator for node and edge records

        """
        if "delimiter" not in kwargs:
            # infer delimiter from file format
            kwargs["delimiter"] = extension_types[format]  # type: ignore
        if "lineterminator" not in kwargs:
            # set '\n' to be the default line terminator to prevent
            # truncation of lines due to hidden/escaped carriage returns
            kwargs["lineterminator"] = "\n"  # type: ignore

        mode = (
            archive_read_mode[compression] if compression in archive_read_mode else None
        )

        self.set_provenance_map(kwargs)

        if format == "tsv":
            kwargs["quoting"] = 3  # type: ignore
        if mode:
            with tarfile.open(filename, mode=mode) as tar:
                # Alas, the order that tar file members is important in some streaming operations
                # (e.g. graph-summary and validation) in that generally, all the node files need to be
                # loaded first,  followed by the  associated edges files can be loaded and analysed.

                # Start by partitioning files of each type into separate lists
                node_files: List[str] = list()
                edge_files: List[str] = list()
                for name in tar.getnames():
                    if re.search(f"nodes.{format}", name):
                        node_files.append(name)
                    elif re.search(f"edges.{format}", name):
                        edge_files.append(name)
                    else:
                        # This used to throw an exception but perhaps we should simply ignore it.
                        log.warning(
                            f"Tar archive contains an unrecognized file: {name}. Skipped..."
                        )

                # Then, first extract and capture contents of the nodes files...
                for name in node_files:
                    try:
                        member = tar.getmember(name)
                    except KeyError:
                        log.warning(
                            f"Node file {name} member in archive {filename} could not be accessed? Skipped?"
                        )
                        continue

                    f = tar.extractfile(member)
                    # TODO: can this somehow be streamed here?
                    file_iter = pd.read_csv(
                        f,
                        dtype=str,
                        chunksize=10000,
                        low_memory=False,
                        keep_default_na=False,
                        **kwargs,
                    )
                    for chunk in file_iter:
                        self.node_properties.update(chunk.columns)
                        yield from self.read_nodes(chunk)

                # Next, extract and capture contents of the edges files...
                for name in edge_files:
                    try:
                        member = tar.getmember(name)
                    except KeyError:
                        log.warning(
                            f"Edge file {name} member in archive {filename} could not be accessed? Skipped?"
                        )
                        continue

                    f = tar.extractfile(member)
                    # TODO: can this somehow be streamed here?
                    file_iter = pd.read_csv(
                        f,
                        dtype=str,
                        chunksize=10000,
                        low_memory=False,
                        keep_default_na=False,
                        **kwargs,
                    )
                    for chunk in file_iter:
                        self.edge_properties.update(chunk.columns)
                        yield from self.read_edges(chunk)
        else:
            file_iter = pd.read_csv(
                filename,
                dtype=str,
                chunksize=10000,
                low_memory=False,
                keep_default_na=False,
                **kwargs,
            )
            if re.search(f"nodes.{format}", filename):
                for chunk in file_iter:
                    self.node_properties.update(chunk.columns)
                    yield from self.read_nodes(chunk)
            elif re.search(f"edges.{format}", filename):
                for chunk in file_iter:
                    self.edge_properties.update(chunk.columns)
                    yield from self.read_edges(chunk)
            else:
                # This used to throw an exception but perhaps we should simply ignore it.
                log.warning(
                    f"Parse function cannot resolve the KGX file type in name {filename}. Skipped..."
                )

    def read_nodes(self, df: pd.DataFrame) -> Generator:
        """
        Read records from pandas.DataFrame and yield records.

        Parameters
        ----------
        df: pandas.DataFrame
            Dataframe containing records that represent nodes

        Returns
        -------
        Generator
            A generator for node records

        """
        for obj in df.to_dict("records"):
            yield self.read_node(obj)

    def read_node(self, node: Dict) -> Optional[Tuple[str, Dict]]:
        """
        Prepare a node.

        Parameters
        ----------
        node: Dict
            A node

        Returns
        -------
        Optional[Tuple[str, Dict]]
            A tuple that contains node id and node data
        """
        node = validate_node(node)
        node_data = sanitize_import(node.copy())
        if "id" in node_data:

            n = node_data["id"]

            self.set_node_provenance(node_data)

            self.node_properties.update(list(node_data.keys()))
            if self.check_node_filter(node_data):
                self.node_properties.update(node_data.keys())
                return n, node_data
        else:
            log.info(f"Ignoring node with no 'id': {node}")

    def read_edges(self, df: pd.DataFrame) -> Generator:
        """
        Load edges from pandas.DataFrame into an instance of BaseGraph.

        Parameters
        ----------
        df: pandas.DataFrame
            Dataframe containing records that represent edges

        Returns
        -------
        Generator
            A generator for edge records

        """
        for obj in df.to_dict("records"):
            yield self.read_edge(obj)

    def read_edge(self, edge: Dict) -> Optional[Tuple]:
        """
        Load an edge into an instance of BaseGraph.

        Parameters
        ----------
        edge: Dict
            An edge

        Returns
        -------
        Optional[Tuple]
            A tuple that contains subject id, object id, edge key, and edge data

        """
        edge = validate_edge(edge)
        edge_data = sanitize_import(edge.copy())
        if "id" not in edge_data:
            edge_data["id"] = generate_uuid()
        s = edge_data["subject"]
        o = edge_data["object"]

        self.set_edge_provenance(edge_data)

        key = generate_edge_key(s, edge_data["predicate"], o)
        self.edge_properties.update(list(edge_data.keys()))
        if self.check_edge_filter(edge_data):
            self.node_properties.update(edge_data.keys())
            return s, o, key, edge_data



import gzip
from typing import Set, Dict, Union, Optional, Any, Tuple, List, Generator

import rdflib
from linkml_runtime.linkml_model.meta import SlotDefinition, ClassDefinition, Element
from rdflib import URIRef, RDF, Namespace

from kgx.prefix_manager import PrefixManager
from kgx.config import get_logger
from kgx.parsers.ntriples_parser import CustomNTriplesParser
from kgx.source.source import Source
from kgx.utils.graph_utils import curie_lookup
from kgx.utils.kgx_utils import (
    get_toolkit,
    get_biolink_property_types,
    is_property_multivalued,
    generate_edge_key,
    sentencecase_to_snakecase,
    sentencecase_to_camelcase,
    get_biolink_ancestors,
    validate_edge,
    validate_node,
    sanitize_import,
    prepare_data_dict,
    DEFAULT_EDGE_PREDICATE,
    CORE_NODE_PROPERTIES,
    CORE_EDGE_PROPERTIES,
    knowledge_provenance_properties,
)

log = get_logger()

NAMED_THING = "biolink:NamedThing"


class RdfSource(Source):
    """
    RdfSource is responsible for reading data as records
    from RDF.

    .. note::
        Currently only RDF N-Triples are supported.

    """

    def __init__(self):
        super().__init__()
        self.DEFAULT = Namespace(self.prefix_manager.prefix_map[""])
        # TODO: use OBO IRI from biolink model context once
        #  https://github.com/biolink/biolink-model/issues/211 is resolved
        # self.OBO = Namespace('http://purl.obolibrary.org/obo/')
        self.OBAN = Namespace(self.prefix_manager.prefix_map["OBAN"])
        self.PMID = Namespace(self.prefix_manager.prefix_map["PMID"])
        self.BIOLINK = Namespace(self.prefix_manager.prefix_map["biolink"])
        self.predicate_mapping = {}
        self.cache: Dict = {}
        self.toolkit = get_toolkit()
        self.node_property_predicates = set(
            [
                URIRef(self.prefix_manager.expand(x))
                for x in self.toolkit.get_all_node_properties(formatted=True)
            ]
        )
        self.node_property_predicates.update(
            set(self.toolkit.get_all_node_properties(formatted=True))
        )
        self.node_property_predicates.update(
            set(self.toolkit.get_all_edge_properties(formatted=True))
        )

        # TODO: validate expansion of the scope of this statement to include 'knowledge_source' and its descendants?
        for ksf in knowledge_provenance_properties:
            self.node_property_predicates.add(
                URIRef(self.prefix_manager.expand("biolink:" + ksf))
            )

        self.reification_types = {
            RDF.Statement,
            self.BIOLINK.Association,
            self.OBAN.association,
        }
        self.reification_predicates = {
            self.BIOLINK.subject,
            self.BIOLINK.predicate,
            self.BIOLINK.object,
            RDF.subject,
            RDF.object,
            RDF.predicate,
            self.OBAN.association_has_subject,
            self.OBAN.association_has_predicate,
            self.OBAN.association_has_object,
        }
        self.reified_nodes: Set = set()
        self.start: int = 0
        self.count: int = 0
        self.CACHE_SIZE = 10000
        self.node_record = {}
        self.edge_record = {}
        self.node_cache = {}
        self.edge_cache = {}
        self._incomplete_nodes = {}

    def set_predicate_mapping(self, m: Dict) -> None:
        """
        Set predicate mappings.

        Use this method to update mappings for predicates that are
        not in Biolink Model.

        Parameters
        ----------
        m: Dict
            A dictionary where the keys are IRIs and values are their
            corresponding property names

        """
        for k, v in m.items():
            self.predicate_mapping[URIRef(k)] = v

    def set_node_property_predicates(self, predicates) -> None:
        """
        Set predicates that are to be treated as node properties.

        Parameters
        ----------
        predicates: Set
            Set of predicates

        """
        for p in predicates:
            self.node_property_predicates.add(URIRef(p))

    def parse(
        self,
        filename: str,
        format: str = "nt",
        compression: Optional[str] = None,
        **kwargs: Any,
    ) -> Generator:
        """
        This method reads from RDF N-Triples and yields records.

        .. note::
            To ensure proper parsing of N-Triples and a relatively low memory footprint,
            it is recommended that the N-Triples be sorted based on the subject IRIs.

            ```sort -k 1,2 -t ' ' data.nt > data_sorted.nt```

        Parameters
        ----------
        filename: str
            The filename to parse
        format: str
            The format (``nt``)
        compression: Optional[str]
            The compression type (``gz``)
        kwargs: Any
            Any additional arguments

        Returns
        -------
        Generator
            A generator for records

        """
        p = CustomNTriplesParser(self)

        self.set_provenance_map(kwargs)

        if compression == "gz":
            yield from p.parse(gzip.open(filename, "rb"))
        else:
            yield from p.parse(open(filename, "rb"))
        log.info(f"Done parsing {filename}")

        for n in self.reified_nodes:
            data = self.node_cache.pop(n)
            self.dereify(n, data)

        for k in self.node_cache.keys():
            node_data = self.node_cache[k]
            if "category" in node_data:
                if NAMED_THING not in set(node_data["category"]):
                    node_data["category"].append(NAMED_THING)
            else:
                node_data["category"] = [NAMED_THING]
            node_data = validate_node(node_data)
            node_data = sanitize_import(node_data)

            self.set_node_provenance(node_data)

            if self.check_node_filter(node_data):
                self.node_properties.update(node_data.keys())
                yield k, node_data

        self.node_cache.clear()

        for k in self.edge_cache.keys():
            edge_data = self.edge_cache[k]
            edge_data = validate_edge(edge_data)
            edge_data = sanitize_import(edge_data)

            self.set_edge_provenance(edge_data)

            if self.check_edge_filter(edge_data):
                self.edge_properties.update(edge_data.keys())
                yield k[0], k[1], k[2], edge_data

        self.edge_cache.clear()

    def triple(self, s: URIRef, p: URIRef, o: URIRef) -> None:
        """
        Parse a triple.

        Parameters
        ----------
        s: URIRef
            Subject
        p: URIRef
            Predicate
        o: URIRef
            Object

        """
        self.count += 1
        (element_uri, canonical_uri, predicate, property_name) = self.process_predicate(
            p
        )
        if element_uri:
            prop_uri = element_uri
        elif predicate:
            prop_uri = predicate
        else:
            prop_uri = property_name

        s_curie = self.prefix_manager.contract(s)
        if s_curie.startswith("biolink") or s_curie.startswith("OBAN"):
            log.warning(f"Skipping {s} {p} {o}")
        elif s_curie in self.reified_nodes:
            # subject is a reified node
            self.add_node_attribute(s, key=prop_uri, value=o)
        elif p in self.reification_predicates:
            # subject is a reified node
            self.reified_nodes.add(s_curie)
            self.add_node_attribute(s, key=prop_uri, value=o)
        elif property_name in {
            "subject",
            "predicate",
            "object",
            "predicate",
            "relation",
        }:
            # subject is a reified node
            self.reified_nodes.add(s_curie)
            self.add_node_attribute(s, key=prop_uri, value=o)
        elif o in self.reification_types:
            # subject is a reified node
            self.reified_nodes.add(s_curie)
            self.add_node_attribute(s, key=prop_uri, value=o)
        elif element_uri and element_uri in self.node_property_predicates:
            # treating predicate as a node property
            self.add_node_attribute(s, key=prop_uri, value=o)
        elif (
            p in self.node_property_predicates
            or predicate in self.node_property_predicates
            or property_name in self.node_property_predicates
        ):
            # treating predicate as a node property
            self.add_node_attribute(s, key=prop_uri, value=o)
        elif isinstance(o, rdflib.term.Literal):
            self.add_node_attribute(s, key=prop_uri, value=o)
        else:
            # treating predicate as an edge
            self.add_edge(s, o, p)

        if len(self.edge_cache) >= self.CACHE_SIZE:
            while self.reified_nodes:
                n = self.reified_nodes.pop()
                data = self.node_cache.pop(n)
                try:
                    self.dereify(n, data)
                except ValueError as e:
                    log.info(e)
                    self._incomplete_nodes[n] = data

            for n in self._incomplete_nodes.keys():
                self.node_cache[n] = self._incomplete_nodes[n]
                self.reified_nodes.add(n)
            self._incomplete_nodes.clear()

            for k in self.edge_cache.keys():
                if (
                    "id" not in self.edge_cache[k]
                    and "association_id" not in self.edge_cache[k]
                ):
                    edge_key = generate_edge_key(
                        self.edge_cache[k]["subject"],
                        self.edge_cache[k]["predicate"],
                        self.edge_cache[k]["object"],
                    )
                    self.edge_cache[k]["id"] = edge_key
                data = self.edge_cache[k]
                data = validate_edge(data)
                data = sanitize_import(data)

                self.set_edge_provenance(data)

                if self.check_edge_filter(data):
                    self.edge_properties.update(data.keys())
                    yield k[0], k[1], k[2], data
            self.edge_cache.clear()
        yield None

    def dereify(self, n: str, node: Dict) -> None:
        """
        Dereify a node to create a corresponding edge.

        Parameters
        ----------
        n: str
            Node identifier
        node: Dict
            Node data

        """
        if "predicate" not in node:
            node["predicate"] = "biolink:related_to"
        if "relation" not in node:
            node["relation"] = node["predicate"]
        # if 'category' in node:
        #     del node['category']
        if "subject" in node and "object" in node:
            self.edge_properties.update(node.keys())
            self.add_edge(node["subject"], node["object"], node["predicate"], node)
        else:
            log.warning(
                f"Missing 'subject' or 'object' in reified edge node {n} {node}. Ignoring the edge...."
            )

    def add_node_attribute(
        self, iri: Union[URIRef, str], key: str, value: Union[str, List]
    ) -> None:
        """
        Add an attribute to a node in cache, while taking into account whether the attribute
        should be multi-valued.

        The ``key`` may be a rdflib.URIRef or an URI string that maps onto a
        property name as defined in ``rdf_utils.property_mapping``.

        Parameters
        ----------
        iri: Union[rdflib.URIRef, str]
            The IRI of a node in the rdflib.Graph
        key: str
            The name of the attribute. Can be a rdflib.URIRef or URI string
        value: Union[str, List]
            The value of the attribute

        Returns
        -------
        Dict
            The node data

        """
        if self.prefix_manager.is_iri(key):
            key_curie = self.prefix_manager.contract(key)
        else:
            key_curie = key
        c = curie_lookup(key_curie)
        if c:
            key_curie = c

        if self.prefix_manager.is_curie(key_curie):
            # property names will always be just the reference
            mapped_key = self.prefix_manager.get_reference(key_curie)
        else:
            mapped_key = key_curie

        if isinstance(value, rdflib.term.Identifier):
            if isinstance(value, rdflib.term.URIRef):
                value_curie = self.prefix_manager.contract(value)
                # if self.prefix_manager.get_prefix(value_curie) not in {'biolink'} \
                #         and mapped_key not in {'type', 'category', 'predicate', 'relation', 'predicate'}:
                #     d = self.add_node(value)
                #     value = d['id']
                # else:
                #     value = value_curie
                value = value_curie
            else:
                value = value.toPython()
        if (
            mapped_key in is_property_multivalued
            and is_property_multivalued[mapped_key]
        ):
            value = [value]
        if mapped_key in self.node_record:
            if isinstance(self.node_record[mapped_key], str):
                _ = self.node_record[mapped_key]
                self.node_record[mapped_key] = [_]
            self.node_record[mapped_key].append(value)
        else:
            self.node_record[mapped_key] = [value]
        curie = self.prefix_manager.contract(iri)
        if curie in self.node_cache:
            if mapped_key in self.node_cache[curie]:
                node = self.node_cache[curie]
                updated_node = prepare_data_dict(node, {mapped_key: value})
                self.node_cache[curie] = updated_node
            else:
                self.node_cache[curie][mapped_key] = value
        else:
            self.node_cache[curie] = {"id": curie, mapped_key: value}

    def add_node(self, iri: URIRef, data: Optional[Dict] = None) -> Dict:
        """
        Add a node to cache.

        Parameters
        ----------
        iri: rdflib.URIRef
            IRI of a node
        data: Optional[Dict]
            Additional node properties

        Returns
        -------
        Dict
            The node data

        """
        n = self.prefix_manager.contract(str(iri))
        if n == iri:
            if self.prefix_manager.has_urlfragment(iri):
                n = rdflib.namespace.urldefrag(iri).fragment

        if not n:
            n = iri
        if n in self.node_cache:
            node_data = self.update_node(n, data)
        else:
            if data:
                node_data = data
            else:
                node_data = {}
            node_data["id"] = n

        if "category" in node_data:
            if NAMED_THING not in set(node_data["category"]):
                node_data["category"].append(NAMED_THING)
        else:
            node_data["category"] = [NAMED_THING]

        self.set_node_provenance(node_data)

        self.node_cache[n] = node_data
        return node_data

    def add_edge(
        self,
        subject_iri: URIRef,
        object_iri: URIRef,
        predicate_iri: URIRef,
        data: Optional[Dict[Any, Any]] = None,
    ) -> Dict:
        """
        Add an edge to cache.

        Parameters
        ----------
        subject_iri: rdflib.URIRef
            Subject IRI for the subject in a triple
        object_iri: rdflib.URIRef
            Object IRI for the object in a triple
        predicate_iri: rdflib.URIRef
            Predicate IRI for the predicate in a triple
        data: Optional[Dict[Any, Any]]
            Additional edge properties

        Returns
        -------
        Dict
            The edge data

        """
        (element_uri, canonical_uri, predicate, property_name) = self.process_predicate(
            predicate_iri
        )
        subject_curie = self.prefix_manager.contract(subject_iri)
        object_curie = self.prefix_manager.contract(object_iri)
        if subject_curie in self.node_cache:
            subject_node = self.node_cache[subject_curie]
        else:
            subject_node = self.add_node(subject_iri)

        if object_curie in self.node_cache:
            object_node = self.node_cache[object_curie]
        else:
            object_node = self.add_node(object_iri)
        edge_predicate = element_uri if element_uri else predicate
        if not edge_predicate:
            edge_predicate = property_name

        if " " in edge_predicate:
            log.debug(
                f"predicate IRI '{predicate_iri}' yields edge_predicate '{edge_predicate}' that not in snake_case form; replacing ' ' with '_'"
            )
        edge_predicate_prefix = self.prefix_manager.get_prefix(edge_predicate)
        if edge_predicate_prefix not in {"biolink", "rdf", "rdfs", "skos", "owl"}:
            if PrefixManager.is_curie(edge_predicate):
                # name = curie_lookup(edge_predicate)
                # if name:
                #     log.debug(f"predicate IRI '{predicate_iri}' yields edge_predicate '{edge_predicate}' that is actually a CURIE; Using its mapping instead: {name}")
                #     edge_predicate = f"{edge_predicate_prefix}:{name}"
                # else:
                #     log.debug(f"predicate IRI '{predicate_iri}' yields edge_predicate '{edge_predicate}' that is actually a CURIE; defaulting back to {self.DEFAULT_EDGE_PREDICATE}")
                edge_predicate = DEFAULT_EDGE_PREDICATE

        edge_key = generate_edge_key(
            subject_node["id"], edge_predicate, object_node["id"]
        )
        if (subject_node["id"], object_node["id"], edge_key) in self.edge_cache:
            # edge already exists; process kwargs and update the edge
            edge_data = self.update_edge(
                subject_node["id"], object_node["id"], edge_key, data
            )
        else:
            # add a new edge
            edge_data = data if data else {}
            edge_data.update(
                {
                    "subject": subject_node["id"],
                    "predicate": f"{edge_predicate}",
                    "object": object_node["id"],
                }
            )
            if "relation" not in edge_data:
                edge_data["relation"] = predicate

            self.set_edge_provenance(edge_data)

        self.edge_cache[(subject_node["id"], object_node["id"], edge_key)] = edge_data
        return edge_data

    def process_predicate(self, p: Optional[Union[URIRef, str]]) -> Tuple:
        """
        Process a predicate where the method checks if there is a mapping in Biolink Model.

        Parameters
        ----------
        p: Optional[Union[URIRef, str]]
            The predicate

        Returns
        -------
        Tuple
            A tuple that contains the Biolink CURIE (if available), the Biolink slot_uri CURIE (if available),
            the CURIE form of p, the reference of p

        """
        if p in self.cache:
            # already processed this predicate before; pull from cache
            element_uri = self.cache[p]["element_uri"]
            canonical_uri = self.cache[p]["canonical_uri"]
            predicate = self.cache[p]["predicate"]
            property_name = self.cache[p]["property_name"]
        else:
            # haven't seen this property before; map to element
            if self.prefix_manager.is_iri(p):
                predicate = self.prefix_manager.contract(str(p))
            else:
                predicate = None
            if self.prefix_manager.is_curie(p):
                property_name = self.prefix_manager.get_reference(p)
                predicate = p
            else:
                if predicate and self.prefix_manager.is_curie(predicate):
                    property_name = self.prefix_manager.get_reference(predicate)
                else:
                    property_name = p
                    predicate = f":{p}"
            element = self.get_biolink_element(p)
            canonical_uri = None
            if element:
                if isinstance(element, SlotDefinition):
                    # predicate corresponds to a biolink slot
                    if element.definition_uri:
                        element_uri = self.prefix_manager.contract(
                            element.definition_uri
                        )
                    else:
                        element_uri = (
                            f"biolink:{sentencecase_to_snakecase(element.name)}"
                        )
                    if element.slot_uri:
                        canonical_uri = element.slot_uri
                elif isinstance(element, ClassDefinition):
                    # this will happen only when the IRI is actually
                    # a reference to a class
                    element_uri = self.prefix_manager.contract(element.class_uri)
                else:
                    element_uri = f"biolink:{sentencecase_to_camelcase(element.name)}"
                if "biolink:Attribute" in get_biolink_ancestors(element.name):
                    element_uri = f"biolink:{sentencecase_to_snakecase(element.name)}"
                if not predicate:
                    predicate = element_uri
            else:
                # no mapping to biolink model;
                # look at predicate mappings
                element_uri = None
                if p in self.predicate_mapping:
                    property_name = self.predicate_mapping[p]
                    predicate = f":{property_name}"
            self.cache[p] = {
                "element_uri": element_uri,
                "canonical_uri": canonical_uri,
                "predicate": predicate,
                "property_name": property_name,
            }
        return element_uri, canonical_uri, predicate, property_name

    def update_node(self, n: Union[URIRef, str], data: Optional[Dict] = None) -> Dict:
        """
        Update a node with properties.

        Parameters
        ----------
        n: Union[URIRef, str]
            Node identifier
        data: Optional[Dict]
            Node properties

        Returns
        -------
        Dict
            The node data

        """
        node_data = self.node_cache[str(n)]
        if data:
            new_data = self._prepare_data_dict(node_data, data)
            node_data.update(new_data)
        return node_data

    def update_edge(
        self,
        subject_curie: str,
        object_curie: str,
        edge_key: str,
        data: Optional[Dict[Any, Any]],
    ) -> Dict:
        """
        Update an edge with properties.

        Parameters
        ----------
        subject_curie: str
            Subject CURIE
        object_curie: str
            Object CURIE
        edge_key: str
            Edge key
        data: Optional[Dict[Any, Any]]
            Edge properties

        Returns
        -------
        Dict
            The edge data

        """
        key = (subject_curie, object_curie, edge_key)
        if key in self.edge_cache:
            edge_data = self.edge_cache[key]
        else:
            edge_data = {}
        if data:
            new_data = self._prepare_data_dict(edge_data, data)
            edge_data.update(new_data)
        return edge_data

    def _prepare_data_dict(self, d1: Dict, d2: Dict) -> Dict:
        """
        Given two dict objects, make a new dict object that is the intersection of the two.

        If a key is known to be multivalued then it's value is converted to a list.
        If a key is already multivalued then it is updated with new values.
        If a key is single valued, and a new unique value is found then the existing value is
        converted to a list and the new value is appended to this list.

        Parameters
        ----------
        d1: Dict
            Dict object
        d2: Dict
            Dict object

        Returns
        -------
        Dict
            The intersection of d1 and d2

        """
        new_data = {}
        for key, value in d2.items():
            if isinstance(value, (list, set, tuple)):
                new_value = [
                    self.prefix_manager.contract(x)
                    if self.prefix_manager.is_iri(x)
                    else x
                    for x in value
                ]
            else:
                new_value = (
                    self.prefix_manager.contract(value)
                    if self.prefix_manager.is_iri(value)
                    else value
                )

            if key in is_property_multivalued:
                if is_property_multivalued[key]:
                    # key is supposed to be multivalued
                    if key in d1:
                        # key is in data
                        if isinstance(d1[key], list):
                            # existing key has value type list
                            new_data[key] = d1[key]
                            if isinstance(new_value, (list, set, tuple)):
                                new_data[key] += [
                                    x for x in new_value if x not in new_data[key]
                                ]
                            else:
                                if new_value not in new_data[key]:
                                    new_data[key].append(new_value)
                        else:
                            if (
                                key in CORE_NODE_PROPERTIES
                                or key in CORE_EDGE_PROPERTIES
                            ):
                                log.debug(
                                    f"cannot modify core property '{key}': {d2[key]} vs {d1[key]}"
                                )
                            else:
                                # existing key does not have value type list; converting to list
                                new_data[key] = [d1[key]]
                                if isinstance(new_value, (list, set, tuple)):
                                    new_data[key] += [
                                        x for x in new_value if x not in new_data[key]
                                    ]
                                else:
                                    if new_value not in new_data[key]:
                                        new_data[key].append(new_value)
                    else:
                        # key is not in data; adding
                        if isinstance(new_value, (list, set, tuple)):
                            new_data[key] = [x for x in new_value]
                        else:
                            new_data[key] = [new_value]
                else:
                    # key is not multivalued; adding/replacing as-is
                    if key in d1:
                        if isinstance(d1[key], list):
                            new_data[key] = d1[key]
                            if isinstance(new_value, (list, set, tuple)):
                                new_data[key] += [x for x in new_value]
                            else:
                                new_data[key].append(new_value)
                        else:
                            if (
                                key in CORE_NODE_PROPERTIES
                                or key in CORE_EDGE_PROPERTIES
                            ):
                                log.debug(
                                    f"cannot modify core property '{key}': {d2[key]} vs {d1[key]}"
                                )
                            else:
                                new_data[key] = new_value
                    else:
                        new_data[key] = new_value
            else:
                # treating key as multivalued
                if key in d1:
                    # key is in data
                    if key in CORE_NODE_PROPERTIES or key in CORE_EDGE_PROPERTIES:
                        log.debug(
                            f"cannot modify core property '{key}': {d2[key]} vs {d1[key]}"
                        )
                    else:
                        if isinstance(d1[key], list):
                            # existing key has value type list
                            new_data[key] = d1[key]
                            if isinstance(new_value, (list, set, tuple)):
                                new_data[key] += [
                                    x for x in new_value if x not in new_data[key]
                                ]
                            else:
                                new_data[key].append(new_value)
                        else:
                            # existing key does not have value type list; converting to list
                            new_data[key] = [d1[key]]
                            if isinstance(new_value, (list, set, tuple)):
                                new_data[key] += [
                                    x for x in new_value if x not in new_data[key]
                                ]
                            else:
                                new_data[key].append(new_value)
                else:
                    new_data[key] = new_value
        return new_data

    def get_biolink_element(self, predicate: Any) -> Optional[Element]:
        """
        Returns a Biolink Model element for a given predicate.

        Parameters
        ----------
        predicate: Any
            The CURIE of a predicate

        Returns
        -------
        Optional[Element]
            The corresponding Biolink Model element

        """
        toolkit = get_toolkit()
        if self.prefix_manager.is_iri(predicate):
            predicate_curie = self.prefix_manager.contract(predicate)
        else:
            predicate_curie = predicate
        if self.prefix_manager.is_curie(predicate_curie):
            reference = self.prefix_manager.get_reference(predicate_curie)
        else:
            reference = predicate_curie
        element = toolkit.get_element(reference)
        if not element:
            try:
                mapping = toolkit.get_element_by_mapping(predicate)
                if mapping:
                    element = toolkit.get_element(mapping)
            except ValueError as e:
                log.error(e)
        return element



import importlib

import os
from os.path import dirname, abspath

import sys
from multiprocessing import Pool
from typing import List, Tuple, Optional, Dict, Set, Any, Union
import yaml

from kgx.validator import Validator
from kgx.sink import Sink
from kgx.transformer import Transformer, SOURCE_MAP, SINK_MAP
from kgx.config import get_logger
from kgx.graph.base_graph import BaseGraph
from kgx.graph_operations.graph_merge import merge_all_graphs
from kgx.graph_operations import summarize_graph, meta_knowledge_graph
from kgx.utils.kgx_utils import apply_graph_operations, knowledge_provenance_properties


summary_report_types = {
    "kgx-map": summarize_graph.GraphSummary,
    "meta-knowledge-graph": meta_knowledge_graph.MetaKnowledgeGraph,
}

log = get_logger()


def get_input_file_types() -> Tuple:
    """
    Get all input file formats supported by KGX.

    Returns
    -------
    Tuple
        A tuple of supported file formats

    """
    return tuple(SOURCE_MAP.keys())


def get_output_file_types() -> Tuple:
    """
    Get all output file formats supported by KGX.

    Returns
    -------
    Tuple
        A tuple of supported file formats

    """
    return tuple(SINK_MAP.keys())


def get_report_format_types() -> Tuple:
    """
    Get all graph summary report formats supported by KGX.

    Returns
    -------
    Tuple
        A tuple of supported file formats

    """
    return "yaml", "json"


def graph_summary(
    inputs: List[str],
    input_format: str,
    input_compression: Optional[str],
    output: Optional[str],
    report_type: str,
    report_format: Optional[str] = None,
    stream: bool = False,
    graph_name: Optional[str] = None,
    node_facet_properties: Optional[List] = None,
    edge_facet_properties: Optional[List] = None,
    error_log: str = "",
) -> Dict:
    """
    Loads and summarizes a knowledge graph from a set of input files.

    Parameters
    ----------
    inputs: List[str]
        Input file
    input_format: str
        Input file format
    input_compression: Optional[str]
        The input compression type
    output: Optional[str]
        Where to write the output (stdout, by default)
    report_type: str
        The summary report type
    report_format: Optional[str]
        The summary report format file types: 'yaml' or 'json'
    stream: bool
        Whether to parse input as a stream
    graph_name: str
        User specified name of graph being summarized
    node_facet_properties: Optional[List]
        A list of node properties from which to generate counts per value for those properties.
        For example, ``['provided_by']``
    edge_facet_properties: Optional[List]
        A list of edge properties (e.g. knowledge_source tags) to facet on.
        For example, ``['original_knowledge_source', 'aggregator_knowledge_source']``
    error_log: str
        Where to write any graph processing error message (stderr, by default)

    Returns
    -------
    Dict
        A dictionary with the graph stats

    """
    if not graph_name:
        graph_name = "Graph"

    if report_format and report_format not in get_report_format_types():
        raise ValueError(f"report_format must be one of {get_report_format_types()}")

    if report_type in summary_report_types:
        # New design pattern enabling 'stream' processing of statistics on a small memory footprint
        # by injecting an inspector in the Transformer.process() source-to-sink data flow.
        #
        # First, we instantiate the Inspector (generally, a Callable class)...
        #
        inspector = summary_report_types[report_type](
            # ...thus, there is no need to hand the Inspector the graph;
            # rather, the inspector will see the graph data after
            # being injected into the Transformer.transform() workflow
            # graph=transformer.store.graph,
            name=graph_name,
            node_facet_properties=node_facet_properties,
            edge_facet_properties=edge_facet_properties,
            error_log=error_log,
        )
    else:
        raise ValueError(f"report_type must be one of {summary_report_types.keys()}")

    if stream:
        output_args = {
            "format": "null"
        }  # streaming processing throws the graph data away
    else:
        output_args = None

    transformer = Transformer(stream=stream)
    transformer.transform(
        input_args={
            "filename": inputs,
            "format": input_format,
            "compression": input_compression,
        },
        output_args=output_args,
        # ... Second, we inject the Inspector into the transform() call,
        # for the underlying Transformer.process() to use...
        inspector=inspector,
    )

    if output:
        with open(output, "w") as gsr:
            inspector.save(gsr, file_format=report_format)
    else:
        inspector.save(sys.stdout, file_format=report_format)

    # ... Third, we directly return the graph statistics to the caller.
    return inspector.get_graph_summary()


def validate(
    inputs: List[str],
    input_format: str,
    input_compression: Optional[str],
    output: Optional[str],
    stream: bool,
    biolink_release: Optional[str] = None,
) -> List:
    """
    Run KGX validator on an input file to check for Biolink Model compliance.

    Parameters
    ----------
    inputs: List[str]
        Input files
    input_format: str
        The input format
    input_compression: Optional[str]
        The input compression type
    output: Optional[str]
        Path to output file (stdout, by default)
    stream: bool
         Whether to parse input as a stream.
    biolink_release: Optional[str] = None
        SemVer version of Biolink Model Release used for validation (default: latest Biolink Model Toolkit version)
    Returns
    -------
    List
        Returns a list of errors, if any

    """
    # New design pattern enabling 'stream' processing of statistics on a small memory footprint
    # by injecting an inspector in the Transformer.process() source-to-sink data flow.
    #
    # First, we instantiate a Validator() class (converted into a Callable class) as an Inspector ...
    # In the new "Inspector" design pattern, we need to instantiate it before the Transformer.
    #
    Validator.set_biolink_model(biolink_release)

    # Validator assumes the currently set Biolink Release
    validator = Validator()

    if stream:
        transformer = Transformer(stream=stream)

        transformer.transform(
            input_args={
                "filename": inputs,
                "format": input_format,
                "compression": input_compression,
            },
            output_args={
                "format": "null"
            },  # streaming processing throws the graph data away
            # ... Second, we inject the Inspector into the transform() call,
            # for the underlying Transformer.process() to use...
            inspector=validator,
        )
    else:
        # "Classical" non-streaming mode, with click.progressbar
        # but an unfriendly large memory footprint for large graphs

        transformer = Transformer()

        transformer.transform(
            {
                "filename": inputs,
                "format": input_format,
                "compression": input_compression,
            },
        )

        # Slight tweak of classical 'validate' function: that the
        # list of errors are cached internally in the Validator object
        validator.validate(transformer.store.graph)

    if output:
        validator.write_report(open(output, "w"))
    else:
        validator.write_report(sys.stdout)

    # ... Third, we return directly any validation errors to the caller
    return validator.get_errors()


def neo4j_download(
    uri: str,
    username: str,
    password: str,
    output: str,
    output_format: str,
    output_compression: Optional[str],
    stream: bool,
    node_filters: Optional[Tuple] = None,
    edge_filters: Optional[Tuple] = None,
) -> Transformer:
    """
    Download nodes and edges from Neo4j database.

    Parameters
    ----------
    uri: str
        Neo4j URI. For example, https://localhost:7474
    username: str
        Username for authentication
    password: str
        Password for authentication
    output: str
        Where to write the output (stdout, by default)
    output_format: Optional[str]
        The output type (``tsv``, by default)
    output_compression: Optional[str]
        The output compression type
    stream: bool
        Whether to parse input as a stream
    node_filters: Optional[Tuple]
        Node filters
    edge_filters: Optional[Tuple]
        Edge filters

    Returns
    -------
    kgx.Transformer
        The NeoTransformer

    """
    transformer = Transformer(stream=stream)
    transformer.transform(
        {
            "uri": uri,
            "username": username,
            "password": password,
            "format": "neo4j",
            "node_filters": node_filters,
            "edge_filters": edge_filters,
        }
    )

    if not output_format:
        output_format = "tsv"
    transformer.save(
        {"filename": output, "format": output_format, "compression": output_compression}
    )
    return transformer


def neo4j_upload(
    inputs: List[str],
    input_format: str,
    input_compression: Optional[str],
    uri: str,
    username: str,
    password: str,
    stream: bool,
    node_filters: Optional[Tuple] = None,
    edge_filters: Optional[Tuple] = None,
) -> Transformer:
    """
    Upload a set of nodes/edges to a Neo4j database.

    Parameters
    ----------
    inputs: List[str]
        A list of files that contains nodes/edges
    input_format: str
        The input format
    input_compression: Optional[str]
        The input compression type
    uri: str
        The full HTTP address for Neo4j database
    username: str
        Username for authentication
    password: str
        Password for authentication
    stream: bool
        Whether to parse input as a stream
    node_filters: Optional[Tuple]
        Node filters
    edge_filters: Optional[Tuple]
        Edge filters

    Returns
    -------
    kgx.Transformer
        The NeoTransformer

    """
    transformer = Transformer(stream=stream)
    transformer.transform(
        {
            "filename": inputs,
            "format": input_format,
            "compression": input_compression,
            "node_filters": node_filters,
            "edge_filters": edge_filters,
        }
    )
    transformer.save(
        {"uri": uri, "username": username, "password": password, "format": "neo4j"}
    )
    return transformer


def _validate_files(cwd: str, file_paths: List[str], context: str = ""):
    """
    Utility method for resolving file paths
    :param cwd: current working directory for resolving possible relative file path names
    :param file_list: list of file path names to resolve
    :param context: optional source context of of the file list
    :return: resolved list of file paths (as absolute paths)
    """
    resolved_files: List[str] = list()
    for f in file_paths:
        # check if the file exists as an absolute path
        if not os.path.exists(f):
            # otherwise, check if file exists as a path
            # relative to the provided "current working directory"
            f = abspath(cwd + "/" + f)
            if not os.path.exists(f):
                raise FileNotFoundError(
                    f"Filename '{f}' for source '{context}' does not exist!"
                )
        if not os.path.isfile(f):
            raise FileNotFoundError(
                f"Filename '{f}' for source '{context}' is not a file!"
            )
        resolved_files.append(f)
    return resolved_files


def _process_knowledge_source(ksf: str, spec: str) -> Union[str, bool, Tuple]:
    if ksf not in knowledge_provenance_properties:
        log.warning("Unknown Knowledge Source Field: " + ksf + "... ignoring!")
        return False
    else:
        if spec.lower() == "true":
            return True
        elif spec.lower() == "false":
            return False
        else:
            # If a Tuple, expect a comma-delimited string?
            spec_parts = spec.split(",")
            if len(spec_parts) == 1:
                # assumed to be just a default string value for the knowledge source field
                return spec_parts[0]
            else:
                # assumed to be an InfoRes Tuple rewrite specification
                if len(spec_parts) > 3:
                    spec_parts = spec_parts[:2]
                return tuple(spec_parts)


def transform(
    inputs: Optional[List[str]],
    input_format: Optional[str] = None,
    input_compression: Optional[str] = None,
    output: Optional[str] = None,
    output_format: Optional[str] = None,
    output_compression: Optional[str] = None,
    stream: bool = False,
    node_filters: Optional[List[Tuple[str, str]]] = None,
    edge_filters: Optional[List[Tuple[str, str]]] = None,
    transform_config: str = None,
    source: Optional[List] = None,
    knowledge_sources: Optional[List[Tuple[str, str]]] = None,
    # this parameter doesn't get used, but I leave it in
    # for now, in case it signifies an unimplemented concept
    # destination: Optional[List] = None,
    processes: int = 1,
    infores_catalog: Optional[str] = None,
) -> None:
    """
    Transform a Knowledge Graph from one serialization form to another.

    Parameters
    ----------
    inputs: Optional[List[str]]
        A list of files that contains nodes/edges
    input_format: Optional[str]
        The input format
    input_compression: Optional[str]
        The input compression type
    output: Optional[str]
        The output file
    output_format: Optional[str]
        The output format
    output_compression: Optional[str]
        The output compression type
    stream: bool
        Whether to parse input as a stream
    node_filters: Optional[List[Tuple[str, str]]]
        Node input filters
    edge_filters: Optional[List[Tuple[str, str]]]
        Edge input filters
    transform_config: Optional[str]
        The transform config YAML
    source: Optional[List]
        A list of source to load from the YAML
    knowledge_sources: Optional[List[Tuple[str, str]]]
        A list of named knowledge sources with (string, boolean or tuple rewrite) specification
    processes: int
        Number of processes to use
    infores_catalog: Optional[str]
        Optional dump of a TSV file of InfoRes CURIE to
        Knowledge Source mappings (not yet available in transform_config calling mode)

    """
    if transform_config and inputs:
        raise ValueError("Can accept either --transform-config OR inputs, not both")

    output_directory = "output"

    if transform_config:
        # Use the directory within which the 'transform_config' file
        # exists as a 'current working directory' for
        # resolving relative filename paths in the configuration.
        cwd = dirname(transform_config)
        cfg = yaml.load(open(transform_config), Loader=yaml.FullLoader)
        top_level_args = {}
        if "configuration" in cfg:
            top_level_args = prepare_top_level_args(cfg["configuration"])
            if (
                "output_directory" in cfg["configuration"]
                and cfg["configuration"]["output_directory"]
            ):
                output_directory = cfg["configuration"]["output_directory"]
                if not output_directory.startswith(os.path.sep):
                    # relative path
                    output_directory = f"{os.path.abspath(os.path.dirname(transform_config))}{os.path.sep}{output_directory}"

        if not os.path.exists(output_directory):
            os.mkdir(output_directory)

        if not source:
            source = cfg["transform"]["source"].keys()
        for s in source:
            source_properties = cfg["transform"]["source"][s]
            if source_properties["input"]["format"] in get_input_file_types():
                source_properties["input"]["filename"] = _validate_files(
                    cwd=cwd,
                    file_paths=source_properties["input"]["filename"],
                    context=s,
                )

        source_to_parse = {}
        for key, val in cfg["transform"]["source"].items():
            if key in source:
                source_to_parse[key] = val

        results = []
        pool = Pool(processes=processes)
        for k, v in source_to_parse.items():
            log.info(f"Spawning process for '{k}'")
            result = pool.apply_async(
                transform_source,
                (
                    k,
                    v,
                    output_directory,
                    top_level_args["prefix_map"],
                    top_level_args["node_property_predicates"],
                    top_level_args["predicate_mappings"],
                    top_level_args["reverse_prefix_map"],
                    top_level_args["reverse_predicate_mappings"],
                    top_level_args["property_types"],
                    top_level_args["checkpoint"],
                    False,
                    stream,
                ),
            )
            results.append(result)
        pool.close()
        pool.join()
        graphs = [r.get() for r in results]
    else:
        source_dict: Dict = {
            "input": {
                "format": input_format,
                "compression": input_compression,
                "filename": inputs,
                "filters": {
                    "node_filters": node_filters,
                    "edge_filters": edge_filters,
                },
            },
            "output": {
                "format": output_format,
                "compression": output_compression,
                "filename": output,
            },
        }

        if knowledge_sources:
            for ksf, spec in knowledge_sources:
                ksf_spec = _process_knowledge_source(ksf, spec)
                if isinstance(ksf_spec, tuple):
                    if ksf not in source_dict["input"]:
                        source_dict["input"][ksf] = dict()
                    if isinstance(source_dict["input"][ksf], dict):
                        key = ksf_spec[0]
                        source_dict["input"][ksf][key] = ksf_spec
                    else:
                        # Unexpected condition - mixing static values with tuple specified rewrites?
                        raise RuntimeError(
                            "Inconsistent multivalued specifications: make sure that all the  values "
                            + "of the knowledge source tag '"
                            + ksf
                            + "' are all rewrite specifications!"
                        )
                else:
                    source_dict["input"][ksf] = ksf_spec

        name = os.path.basename(inputs[0])
        transform_source(
            key=name,
            source=source_dict,
            output_directory=None,
            stream=stream,
            infores_catalog=infores_catalog,
        )


def merge(
    merge_config: str,
    source: Optional[List] = None,
    destination: Optional[List] = None,
    processes: int = 1,
) -> BaseGraph:
    """
    Load nodes and edges from files and KGs, as defined in a config YAML, and merge them into a single graph.
    The merged graph can then be written to a local/remote Neo4j instance OR be serialized into a file.

    Parameters
    ----------
    merge_config: str
        Merge config YAML
    source: Optional[List]
        A list of source to load from the YAML
    destination: Optional[List]
        A list of destination to write to, as defined in the YAML
    processes: int
        Number of processes to use

    Returns
    -------
    kgx.graph.base_graph.BaseGraph
        The merged graph

    """
    # Use the directory within which the 'merge_config' file
    # exists as a 'current working directory' for
    # resolving relative filename paths in the configuration.
    cwd = dirname(merge_config)

    with open(merge_config, "r") as YML:
        cfg = yaml.load(YML, Loader=yaml.FullLoader)

    output_directory = "output"

    top_level_args = {}
    if "configuration" in cfg:
        top_level_args = prepare_top_level_args(cfg["configuration"])
        if (
            "output_directory" in cfg["configuration"]
            and cfg["configuration"]["output_directory"]
        ):
            output_directory = cfg["configuration"]["output_directory"]
            if not output_directory.startswith(os.path.sep):
                # relative path
                output_directory = f"{os.path.abspath(os.path.dirname(merge_config))}{os.path.sep}{output_directory}"

    if not os.path.exists(output_directory):
        os.mkdir(output_directory)

    if not source:
        source = cfg["merged_graph"]["source"].keys()

    if not destination:
        destination = cfg["merged_graph"]["destination"].keys()

    for s in source:
        source_properties = cfg["merged_graph"]["source"][s]
        if source_properties["input"]["format"] in get_input_file_types():
            source_properties["input"]["filename"] = _validate_files(
                cwd=cwd, file_paths=source_properties["input"]["filename"], context=s
            )

    sources_to_parse = {}
    for key in cfg["merged_graph"]["source"]:
        if key in source:
            sources_to_parse[key] = cfg["merged_graph"]["source"][key]

    results = []
    pool = Pool(processes=processes)
    for k, v in sources_to_parse.items():
        log.info(f"Spawning process for '{k}'")
        result = pool.apply_async(
            parse_source,
            (
                k,
                v,
                output_directory,
                top_level_args["prefix_map"],
                top_level_args["node_property_predicates"],
                top_level_args["predicate_mappings"],
                top_level_args["checkpoint"],
            ),
        )
        results.append(result)
    pool.close()
    pool.join()
    stores = [r.get() for r in results]
    merged_graph = merge_all_graphs([x.graph for x in stores])
    log.info(
        f"Merged graph has {merged_graph.number_of_nodes()} nodes and {merged_graph.number_of_edges()} edges"
    )
    if "name" in cfg["merged_graph"]:
        merged_graph.name = cfg["merged_graph"]["name"]
    if "operations" in cfg["merged_graph"]:
        apply_graph_operations(merged_graph, cfg["merged_graph"]["operations"])

    destination_to_write: Dict[str, Dict] = {}
    for d in destination:
        if d in cfg["merged_graph"]["destination"]:
            destination_to_write[d] = cfg["merged_graph"]["destination"][d]
        else:
            raise KeyError(f"Cannot find destination '{d}' in YAML")

    # write the merged graph
    node_properties = set()
    edge_properties = set()
    for s in stores:
        node_properties.update(s.node_properties)
        edge_properties.update(s.edge_properties)

    input_args = {"graph": merged_graph, "format": "graph"}
    if destination_to_write:
        for key, destination_info in destination_to_write.items():
            log.info(f"Writing merged graph to {key}")
            output_args = {
                "format": destination_info["format"],
                "reverse_prefix_map": top_level_args["reverse_prefix_map"],
                "reverse_predicate_mappings": top_level_args[
                    "reverse_predicate_mappings"
                ],
            }
            if "reverse_prefix_map" in destination_info:
                output_args["reverse_prefix_map"].update(
                    destination_info["reverse_prefix_map"]
                )
            if "reverse_predicate_mappings" in destination_info:
                output_args["reverse_predicate_mappings"].update(
                    destination_info["reverse_predicate_mappings"]
                )
            if destination_info["format"] == "neo4j":
                output_args["uri"] = destination_info["uri"]
                output_args["username"] = destination_info["username"]
                output_args["password"] = destination_info["password"]
            elif destination_info["format"] in get_input_file_types():
                filename = destination_info["filename"]
                if isinstance(filename, list):
                    filename = filename[0]
                destination_filename = f"{output_directory}/{filename}"
                output_args["filename"] = destination_filename
                output_args["compression"] = (
                    destination_info["compression"]
                    if "compression" in destination_info
                    else None
                )
                if destination_info['format'] == 'nt':
                    output_args['property_types'] = top_level_args['property_types']
                    if 'property_types' in top_level_args and 'property_types' in destination_info.keys():
                        output_args['property_types'].update(destination_info['property_types'])
                if destination_info['format'] in {'csv', 'tsv'}:
                    output_args['node_properties'] = node_properties
                    output_args['edge_properties'] = edge_properties
            else:
                raise TypeError(
                    f"type {destination_info['format']} not yet supported for KGX merge operation."
                )
            transformer = Transformer()
            transformer.transform(input_args, output_args)
    else:
        log.warning(
            f"No destination provided in {merge_config}. The merged graph will not be persisted."
        )
    return merged_graph


def parse_source(
    key: str,
    source: dict,
    output_directory: str,
    prefix_map: Dict[str, str] = None,
    node_property_predicates: Set[str] = None,
    predicate_mappings: Dict[str, str] = None,
    checkpoint: bool = False,
) -> Sink:
    """
    Parse a source from a merge config YAML.

    Parameters
    ----------
    key: str
        Source key
    source: Dict
        Source configuration
    output_directory: str
        Location to write output to
    prefix_map: Dict[str, str]
        Non-canonical CURIE mappings
    node_property_predicates: Set[str]
        A set of predicates that ought to be treated as node properties (This is applicable for RDF)
    predicate_mappings: Dict[str, str]
        A mapping of predicate IRIs to property names (This is applicable for RDF)
    checkpoint: bool
        Whether to serialize each individual source to a TSV

    Returns
    -------
    kgx.sink.sink.Sink
        Returns an instance of Sink

    """
    log.info(f"Processing source '{key}'")
    if not key:
        key = os.path.basename(source["input"]["filename"][0])
    input_args = prepare_input_args(
        key,
        source,
        output_directory,
        prefix_map,
        node_property_predicates,
        predicate_mappings,
    )
    transformer = Transformer()
    transformer.transform(input_args)
    transformer.store.graph.name = key
    if checkpoint:
        log.info(f"Writing checkpoint for source '{key}'")
        checkpoint_output = f"{output_directory}/{key}" if output_directory else key
        transformer.save({"filename": checkpoint_output, "format": "tsv"})

    # Current "Callable" metadata not needed at this  point
    # but causes peculiar problems downstream, so we clear it.
    transformer.store.clear_graph_metadata()

    return transformer.store


def transform_source(
    key: str,
    source: Dict,
    output_directory: Optional[str],
    prefix_map: Dict[str, str] = None,
    node_property_predicates: Set[str] = None,
    predicate_mappings: Dict[str, str] = None,
    reverse_prefix_map: Dict = None,
    reverse_predicate_mappings: Dict = None,
    property_types: Dict = None,
    checkpoint: bool = False,
    preserve_graph: bool = True,
    stream: bool = False,
    infores_catalog: Optional[str] = None,
) -> Sink:
    """
    Transform a source from a transform config YAML.

    Parameters
    ----------
    key: str
        Source key
    source: Dict
        Source configuration
    output_directory: Optional[str]
        Location to write output to
    prefix_map: Dict[str, str]
        Non-canonical CURIE mappings
    node_property_predicates: Set[str]
        A set of predicates that ought to be treated as node properties (This is applicable for RDF)
    predicate_mappings: Dict[str, str]
        A mapping of predicate IRIs to property names (This is applicable for RDF)
    reverse_prefix_map: Dict[str, str]
        Non-canonical CURIE mappings for export
    reverse_predicate_mappings: Dict[str, str]
        A mapping of property names to predicate IRIs (This is applicable for RDF)
    property_types: Dict[str, str]
        The xml property type for properties that are other than ``xsd:string``.
        Relevant for RDF export.
    checkpoint: bool
        Whether to serialize each individual source to a TSV
    preserve_graph: true
        Whether or not to preserve the graph corresponding to the source
    stream: bool
        Whether to parse input as a stream
    infores_catalog: Optional[str]
        Optional dump of a TSV file of InfoRes CURIE to Knowledge Source mappings

    Returns
    -------
    kgx.sink.sink.Sink
        Returns an instance of Sink

    """
    log.info(f"Processing source '{key}'")
    input_args = prepare_input_args(
        key,
        source,
        output_directory,
        prefix_map,
        node_property_predicates,
        predicate_mappings,
    )
    output_args = prepare_output_args(
        key,
        source,
        output_directory,
        reverse_prefix_map,
        reverse_predicate_mappings,
        property_types,
    )
    transformer = Transformer(stream=stream, infores_catalog=infores_catalog)
    transformer.transform(input_args, output_args)

    if not preserve_graph:
        transformer.store.graph.clear()

    if infores_catalog:
        with open(infores_catalog, "w") as irc:
            catalog: Dict[str, str] = transformer.get_infores_catalog()
            for source in catalog.keys():
                infores = catalog.setdefault(source, "unknown")
                print(f"{source}\t{infores}", file=irc)

    return transformer.store


def prepare_input_args(
    key: str,
    source: Dict,
    output_directory: Optional[str],
    prefix_map: Dict[str, str] = None,
    node_property_predicates: Set[str] = None,
    predicate_mappings: Dict[str, str] = None,
) -> Dict:
    """
    Prepare input arguments for Transformer.

    Parameters
    ----------
    key: str
        Source key
    source: Dict
        Source configuration
    output_directory: str
        Location to write output to
    prefix_map: Dict[str, str]
        Non-canonical CURIE mappings
    node_property_predicates: Set[str]
        A set of predicates that ought to be treated as node properties (This is applicable for RDF)
    predicate_mappings: Dict[str, str]
        A mapping of predicate IRIs to property names (This is applicable for RDF)

    Returns
    -------
    Dict
        Input arguments as dictionary

    """
    if not key:
        key = os.path.basename(source["input"]["filename"][0])
    input_format = source["input"]["format"]
    input_compression = (
        source["input"]["compression"] if "compression" in source["input"] else None
    )
    inputs = source["input"]["filename"]
    filters = (
        source["input"]["filters"]
        if "filters" in source["input"] and source["input"]["filters"] is not None
        else {}
    )
    node_filters = filters["node_filters"] if "node_filters" in filters else {}
    edge_filters = filters["edge_filters"] if "edge_filters" in filters else {}
    source_prefix_map = prefix_map.copy() if prefix_map else {}
    source_prefix_map.update(
        source["prefix_map"] if "prefix_map" in source and source["prefix_map"] else {}
    )
    source_predicate_mappings = predicate_mappings.copy() if predicate_mappings else {}
    source_predicate_mappings.update(
        source["predicate_mappings"]
        if "predicate_mappings" in source and source["predicate_mappings"] is not None
        else {}
    )
    source_node_property_predicates = (
        node_property_predicates if node_property_predicates else set()
    )
    source_node_property_predicates.update(
        source["node_property_predicates"]
        if "node_property_predicates" in source
        and source["node_property_predicates"] is not None
        else set()
    )

    if input_format in {"nt", "ttl"}:
        input_args = {
            "filename": inputs,
            "format": input_format,
            "compression": input_compression,
            "node_filters": node_filters,
            "edge_filters": edge_filters,
            "prefix_map": source_prefix_map,
            "predicate_mappings": source_predicate_mappings,
            "node_property_predicates": source_node_property_predicates,
        }
    elif input_format in get_input_file_types():
        input_args = {
            "filename": inputs,
            "format": input_format,
            "compression": input_compression,
            "node_filters": node_filters,
            "edge_filters": edge_filters,
            "prefix_map": source_prefix_map,
        }
    elif input_format == "neo4j":
        input_args = {
            "uri": source["uri"],
            "username": source["username"],
            "password": source["password"],
            "format": input_format,
            "node_filters": node_filters,
            "edge_filters": edge_filters,
            "prefix_map": prefix_map,
        }
    else:
        raise TypeError(f"Type {input_format} not yet supported")

    for ksf in knowledge_provenance_properties:
        if ksf in source["input"]:
            input_args[ksf] = source["input"][ksf]

    input_args["operations"] = source["input"].get("operations", [])
    for o in input_args["operations"]:
        args = o["args"]
        if "filename" in args:
            filename = args["filename"]
            if not filename.startswith(output_directory):
                o["args"] = os.path.join(output_directory, filename)
    return input_args


def prepare_output_args(
    key: str,
    source: Dict,
    output_directory: Optional[str],
    reverse_prefix_map: Dict = None,
    reverse_predicate_mappings: Dict = None,
    property_types: Dict = None,
) -> Dict:
    """
    Prepare output arguments for Transformer.

    Parameters
    ----------
    key: str
        Source key
    source: Dict
        Source configuration
    output_directory: str
        Location to write output to
    reverse_prefix_map: Dict[str, str]
        Non-canonical CURIE mappings for export
    reverse_predicate_mappings: Dict[str, str]
        A mapping of property names to predicate IRIs (This is applicable for RDF)
    property_types: Dict[str, str]
        The xml property type for properties that are other than ``xsd:string``.
        Relevant for RDF export.

    Returns
    -------
    Dict
        Output arguments as dictionary

    """
    output_format = source["output"]["format"]
    output_compression = (
        source["output"]["compression"] if "compression" in source["output"] else None
    )
    output_filename = (
        source["output"]["filename"] if "filename" in source["output"] else key
    )
    source_reverse_prefix_map = reverse_prefix_map.copy() if reverse_prefix_map else {}
    source_reverse_prefix_map.update(
        source["reverse_prefix_map"]
        if "reverse_prefix_map" in source and source["reverse_prefix_map"]
        else {}
    )
    source_reverse_predicate_mappings = (
        reverse_predicate_mappings.copy() if reverse_predicate_mappings else {}
    )
    source_reverse_predicate_mappings.update(
        source["reverse_predicate_mappings"]
        if "reverse_predicate_mappings" in source
        and source["reverse_predicate_mappings"] is not None
        else {}
    )
    source_property_types = property_types.copy() if property_types else {}
    source_property_types.update(
        source["property_types"]
    ) if "property_types" in source and source["property_types"] is not None else {}

    if isinstance(output_filename, list):
        output = output_filename[0]
    else:
        output = output_filename
    if output_directory and not output.startswith(output_directory):
        output = os.path.join(output_directory, output)
    output_args = {"format": output_format}
    if output_format == "neo4j":
        output_args["uri"] = source["output"]["uri"]
        output_args["username"] = source["output"]["username"]
        output_args["password"] = source["output"]["password"]
    elif output_format in get_input_file_types():
        output_args["filename"] = output
        output_args["compression"] = output_compression
        if output_format == "nt":
            output_args["reify_all_edges"] = (
                source["output"]["reify_all_edges"]
                if "reify_all_edges" in source["output"]
                else False
            )
            output_args["reverse_prefix_map"] = source_reverse_prefix_map
            output_args[
                "reverse_predicate_mappings"
            ] = source_reverse_predicate_mappings
            output_args["property_types"] = source_property_types
    else:
        raise ValueError(f"type {output_format} not yet supported for output")
    return output_args


def apply_operations(source: dict, graph: BaseGraph) -> BaseGraph:
    """
    Apply operations as defined in the YAML.

    Parameters
    ----------
    source: dict
        The source from the YAML
    graph: kgx.graph.base_graph.BaseGraph
        The graph corresponding to the source

    Returns
    -------
    kgx.graph.base_graph.BaseGraph
        The graph corresponding to the source

    """
    operations = source["operations"]
    for operation in operations:
        op_name = operation["name"]
        op_args = operation["args"]
        module_name = ".".join(op_name.split(".")[0:-1])
        function_name = op_name.split(".")[-1]
        f = getattr(importlib.import_module(module_name), function_name)
        log.info(f"Applying operation {op_name} with args: {op_args}")
        f(graph, **op_args)
    return graph


def prepare_top_level_args(d: Dict) -> Dict:
    """
    Parse top-level configuration.

    Parameters
    ----------
    d: Dict
        The configuration section from the transform/merge YAML

    Returns
    -------
    Dict
        A parsed dictionary with parameters from configuration

    """
    args = {}
    if "checkpoint" in d and d["checkpoint"] is not None:
        args["checkpoint"] = d["checkpoint"]
    else:
        args["checkpoint"] = False
    if "node_property_predicates" in d and d["node_property_predicates"]:
        args["node_property_predicates"] = set(d["node_property_predicates"])
    else:
        args["node_property_predicates"] = set()
    if "predicate_mappings" in d and d["predicate_mappings"]:
        args["predicate_mappings"] = d["predicate_mappings"]
    else:
        args["predicate_mappings"] = {}
    if "prefix_map" in d and d["prefix_map"]:
        args["prefix_map"] = d["prefix_map"]
    else:
        args["prefix_map"] = {}
    if "reverse_prefix_map" in d and d["reverse_prefix_map"] is not None:
        args["reverse_prefix_map"] = d["reverse_prefix_map"]
    else:
        args["reverse_prefix_map"] = {}
    if (
        "reverse_predicate_mappings" in d
        and d["reverse_predicate_mappings"] is not None
    ):
        args["reverse_predicate_mappings"] = d["reverse_predicate_mappings"]
    else:
        args["reverse_predicate_mappings"] = {}
    if "property_types" in d and d["property_types"]:
        args["property_types"] = d["property_types"]
    else:
        args["property_types"] = {}
    return args



from typing import Dict, List, Optional, Any, Callable, Set, Tuple, Union
from sys import stderr

import re

import yaml
from json import dump
from json.encoder import JSONEncoder

from kgx.utils.kgx_utils import GraphEntityType
from kgx.prefix_manager import PrefixManager
from kgx.graph.base_graph import BaseGraph

"""
Generate a knowledge map that corresponds to TRAPI KnowledgeMap.
Specification based on TRAPI Draft PR: https://github.com/NCATSTranslator/ReasonerAPI/pull/171
"""


####################################################################
# Next Generation Implementation of Graph Summary coding which
# leverages the new "Transformer.process()" data stream "Inspector"
# design pattern, implemented here as a "Callable" inspection class.
####################################################################
def mkg_default(o):
    """
    JSONEncoder 'default' function override to
    properly serialize 'Set' objects (into 'List')
    """
    if isinstance(o, MetaKnowledgeGraph.Category):
        return o.json_object()
    else:
        try:
            iterable = iter(o)
        except TypeError:
            pass
        else:
            return list(iterable)
        # Let the base class default method raise the TypeError
        return JSONEncoder().default(o)


_category_curie_regexp = re.compile("^biolink:[A-Z][a-zA-Z]*$")
_predicate_curie_regexp = re.compile("^biolink:[a-z][a-z_]*$")


# DRY parse warning message
def _parse_warning(prefix: str, item: str, suffix: str = None):
    print(
        prefix + " '" + item + "'" + (" " + suffix if suffix else "") + "? Ignoring...",
        file=MetaKnowledgeGraph.error_log,
    )


class MetaKnowledgeGraph:
    """
    Class for generating a TRAPI 1.1 style of "meta knowledge graph" summary.

    The optional 'progress_monitor' for the validator should be a lightweight Callable
    which is injected into the class 'inspector' Callable, designed to intercepts
    node and edge records streaming through the Validator (inside a Transformer.process() call.
    The first (GraphEntityType) argument of the Callable tags the record as a NODE or an EDGE.
    The second argument given to the Callable is the current record itself.
    This Callable is strictly meant to be procedural and should *not* mutate the record.
    The intent of this Callable is to provide a hook to KGX applications wanting the
    namesake function of passively monitoring the graph data stream. As such, the Callable
    could simply tally up the number of times it is called with a NODE or an EDGE, then
    provide a suitable (quick!) report of that count back to the KGX application. The
    Callable (function/callable class) should not modify the record and should be of low
    complexity, so as not to introduce a large computational overhead to validation!
    """

    error_log = stderr

    def __init__(
        self,
        name="",
        node_facet_properties: Optional[List] = None,
        edge_facet_properties: Optional[List] = None,
        progress_monitor: Optional[Callable[[GraphEntityType, List], None]] = None,
        error_log=None,
        **kwargs,
    ):
        """
            MetaKnowledgeGraph constructor.

        Parameters
        ----------
        name: str
            (Graph) name assigned to the summary.
        node_facet_properties: Optional[List]
                A list of node properties (e.g. knowledge_source tags) to facet on. For example, ``['provided_by']``
        edge_facet_properties: Optional[List]
                A list of edge properties (e.g. knowledge_source tags) to facet on. For example,
                ``['original_knowledge_source', 'aggregator_knowledge_source']``
        progress_monitor: Optional[Callable[[GraphEntityType, List], None]]
            Function given a peek at the current record being stream processed by the class wrapped Callable.
        error_log:
            Where to write any graph processing error message (stderr, by default).
        """
        # formal args

        self.name = name

        # these facet properties are used mainly for knowledge_source counting
        # using Biolink 2.0 'knowledge_source' slot values
        if node_facet_properties:
            self.node_facet_properties: Optional[List] = node_facet_properties
        else:
            # node counts still default to 'provided_by'
            self.node_facet_properties: Optional[List] = ["provided_by"]

        if edge_facet_properties:
            self.edge_facet_properties: Optional[List] = edge_facet_properties
        else:
            # node counts still default to 'knowledge_source'
            self.edge_facet_properties: Optional[List] = ["knowledge_source"]

        self.progress_monitor: Optional[
            Callable[[GraphEntityType, List], None]
        ] = progress_monitor

        if error_log:
            MetaKnowledgeGraph.error_log = open(error_log, "w")

        # internal attributes
        # For Nodes...
        self.node_catalog: Dict[str, List[int]] = dict()
        self.node_stats: Dict[str, MetaKnowledgeGraph.Category] = dict()

        # We no longer track 'unknown' categories in meta-knowledge-graph
        # computations since such nodes are not TRAPI 1.1 compliant categories
        # self.node_stats['unknown'] = self.Category('unknown')

        # For Edges...
        self.edge_record_count: int = 0
        self.predicates: Dict = dict()
        self.association_map: Dict = dict()
        self.edge_stats = []

        # Overall graph statistics
        self.graph_stats: Dict[str, Dict] = dict()

    def get_name(self) -> str:
        """
        Returns
        -------
        str
            Currently assigned knowledge graph name.
        """
        return self.name

    def __call__(self, entity_type: GraphEntityType, rec: List):
        """
        Transformer 'inspector' Callable, for analysing a stream of graph data.

        Parameters
        ----------
        entity_type: GraphEntityType
            indicates what kind of record being passed to the function for analysis.
        rec: Dict
            Complete data dictionary of the given record.

        """
        if self.progress_monitor:
            self.progress_monitor(entity_type, rec)
        if entity_type == GraphEntityType.EDGE:
            self.analyse_edge(*rec)
        elif entity_type == GraphEntityType.NODE:
            self.analyse_node(*rec)
        else:
            raise RuntimeError("Unexpected GraphEntityType: " + str(entity_type))

    @staticmethod
    def get_facet_counts(facets: Optional[List], counts_by_source: Dict, data: Dict):
        unknown: bool = True
        for facet in facets:
            if facet in data:
                unknown = False
                if isinstance(data[facet], str):
                    facet_values = [data[facet]]
                else:
                    # assume regular iterable
                    facet_values = list(data[facet])
                for s in facet_values:
                    if facet not in counts_by_source:
                        counts_by_source[facet] = dict()
                    if s in counts_by_source[facet]:
                        counts_by_source[facet][s] += 1
                    else:
                        counts_by_source[facet][s] = 1
        if unknown:
            if "unknown" in counts_by_source:
                counts_by_source["unknown"] += 1
            else:
                counts_by_source["unknown"] = 1

    class Category:
        """
        Internal class for compiling statistics about a distinct category.
        """

        # The 'category map' just associates a unique int catalog
        # index ('cid') value as a proxy for the full curie string,
        # to reduce storage in the main node catalog
        _category_curie_map: List[str] = list()

        def __init__(self, category_curie: str, mkg):
            """
            MetaKnowledgeGraph.Category constructor.

            category_curie: str
                Biolink Model category CURIE identifier.
            """
            if not (
                _category_curie_regexp.fullmatch(category_curie)
                or category_curie == "unknown"
            ):
                raise RuntimeError("Invalid Biolink category CURIE: " + category_curie)

            self.category_curie = category_curie
            self.mkg = mkg

            if self.category_curie not in self._category_curie_map:
                self._category_curie_map.append(self.category_curie)
            self.category_stats: Dict[str, Any] = dict()
            self.category_stats["id_prefixes"] = set()
            self.category_stats["count"] = 0
            self.category_stats["count_by_source"] = dict()

        def get_name(self) -> str:
            """
            Returns
            -------
            str
                CURIE name of the category.
            """
            return self.category_curie

        def get_cid(self):
            """
            Returns
            -------
            int
                Internal MetaKnowledgeGraph index id for tracking a Category.
            """
            return self._category_curie_map.index(self.category_curie)

        @classmethod
        def get_category_curie_from_index(cls, cid: int) -> str:
            """
            Parameters
            ----------
            cid: int
                Internal MetaKnowledgeGraph index id for tracking a Category.

            Returns
            -------
            str
                Curie identifier of the Category.
            """
            return cls._category_curie_map[cid]

        def get_id_prefixes(self) -> Set[str]:
            """
            Returns
            -------
            Set[str]
                Set of identifier prefix (strings) used by nodes of this Category.
            """
            return self.category_stats["id_prefixes"]

        def get_count(self) -> int:
            """
            Returns
            -------
            int
                Count of nodes which have this category.
            """
            return self.category_stats["count"]

        def get_count_by_source(
            self, facet: str = "provided_by", source: str = None
        ) -> Dict[str, Any]:
            """
            Parameters
            ----------
            facet: str
                Facet tag (default, 'provided_by') from which the count should be returned
            source: str
                Source name about which the count is desired.

            Returns
            -------
            Dict
                Count of nodes, by node 'provided_by' knowledge source, for a given category.
                Returns dictionary of all source counts, if input 'source' argument is not specified.
            """
            if source and facet in self.category_stats["count_by_source"]:
                if source in self.category_stats["count_by_source"][facet]:
                    return {
                        source: self.category_stats["count_by_source"][facet][source]
                    }
                else:
                    return {source: 0}
            return self.category_stats["count_by_source"]

        def _compile_prefix_stats(self, n: str):
            prefix = PrefixManager.get_prefix(n)
            if not prefix:
                print(
                    f"Warning: node id {n} has no CURIE prefix",
                    file=MetaKnowledgeGraph.error_log,
                )
            else:
                if prefix not in self.category_stats["id_prefixes"]:
                    self.category_stats["id_prefixes"].add(prefix)

        def _compile_category_source_stats(self, data: Dict):
            self.mkg.get_facet_counts(
                self.mkg.node_facet_properties,
                self.category_stats["count_by_source"],
                data,
            )

        def analyse_node_category(self, n, data) -> None:
            """
            Analyse metadata of a given graph node record of this category.

            Parameters
            ----------
            n: str
                Curie identifier of the node record (not used here).
            data: Dict
                Complete data dictionary of node record fields.

            """
            self.category_stats["count"] += 1
            self._compile_prefix_stats(n)
            self._compile_category_source_stats(data)

        def json_object(self):
            """
            Returns
            -------
            Dict[str, Any]
                Returns JSON friendly metadata for this category.,
            """
            return {
                "id_prefixes": list(self.category_stats["id_prefixes"]),
                "count": self.category_stats["count"],
                "count_by_source": self.category_stats["count_by_source"],
            }

    def get_category(self, category_curie: str) -> Category:
        """
        Counts the number of distinct (Biolink) categories encountered
        in the knowledge graph (not including those of 'unknown' category)

        Parameters
        ----------
        category_curie: str
            Curie identifier for the (Biolink) category.

        Returns
        -------
        Category
            MetaKnowledgeGraph.Category object for a given Biolink category.
        """
        return self.node_stats[category_curie]

    def _process_category_field(self, category_field: str, n: str, data: Dict):
        # we note here that category_curie *may be*
        # a piped '|' set of Biolink category CURIE values
        category_list = category_field.split("|")

        # analyse them each independently...
        for category_curie in category_list:

            if category_curie not in self.node_stats:
                try:
                    self.node_stats[category_curie] = self.Category(
                        category_curie, self
                    )
                except RuntimeError:
                    _parse_warning("Invalid category CURIE", category_curie)
                    continue

            category_record = self.node_stats[category_curie]

            category_idx: int = category_record.get_cid()
            if category_idx not in self.node_catalog[n]:
                self.node_catalog[n].append(category_idx)

            category_record.analyse_node_category(n, data)

    def analyse_node(self, n: str, data: Dict) -> None:
        """
        Analyse metadata of one graph node record.

        Parameters
        ----------
        n: str
            Curie identifier of the node record (not used here).
        data: Dict
            Complete data dictionary of node record fields.

        """
        # The TRAPI release 1.1 meta_knowledge_graph format indexes nodes by biolink:Category
        # the node 'category' field is a list of assigned categories (usually just one...).
        # However, this may perhaps sometimes result in duplicate counting and conflation of prefixes(?).
        if n in self.node_catalog:
            # Report duplications of node records, as discerned from node id.
            _parse_warning(
                "Duplicate node identifier", n, "encountered in input node data?"
            )
            return
        else:
            self.node_catalog[n] = list()

        if "category" not in data or not data["category"]:
            # we now simply exclude nodes with missing categories from the count, since a category
            # of 'unknown' in the  meta_knowledge_graph output  is considered invalid.
            # category = self.node_stats['unknown']
            # category.analyse_node_category(n, data)
            print(
                "Node with identifier '"
                + n
                + "' is missing its 'category' value? Ignoring in the analysis...",
                file=MetaKnowledgeGraph.error_log,
            )
            return

        categories = data["category"]

        # analyse them each independently...
        for category_field in categories:
            self._process_category_field(category_field, n, data)

    def _capture_predicate(self, data: Dict) -> Optional[str]:
        if "predicate" not in data:
            # We no longer track edges with 'unknown' predicates,
            # since those would not be TRAPI 1.1 JSON compliant...
            # self.predicates['unknown'] += 1
            # predicate = "unknown"
            _parse_warning("Empty predicate CURIE in edge data", str(data))
            self.edge_record_count -= 1
            return None
        else:
            predicate = data["predicate"]

            if not _predicate_curie_regexp.fullmatch(predicate):
                _parse_warning("Invalid predicate CURIE", predicate)
                self.edge_record_count -= 1
                return None

            if predicate not in self.predicates:
                # just need to track the number
                # of edge records using this predicate
                self.predicates[predicate] = 0
            self.predicates[predicate] += 1

        return predicate

    def _compile_triple_source_stats(self, triple: Tuple[str, str, str], data: Dict):
        self.get_facet_counts(
            self.edge_facet_properties,
            self.association_map[triple]["count_by_source"],
            data,
        )

    def _process_triple(
        self, subject_category: str, predicate: str, object_category: str, data: Dict
    ):
        # Process the 'valid' S-P-O triple here...
        triple = (subject_category, predicate, object_category)
        if triple not in self.association_map:
            self.association_map[triple] = {
                "subject": triple[0],
                "predicate": triple[1],
                "object": triple[2],
                "relations": set(),
                "count_by_source": dict(),
                "count": 0,
            }

        if (
            "relation" in data
            and data["relation"] not in self.association_map[triple]["relations"]
        ):
            self.association_map[triple]["relations"].add(data["relation"])

        self.association_map[triple]["count"] += 1

        self._compile_triple_source_stats(triple, data)

    def analyse_edge(self, u, v, k, data) -> None:
        """
        Analyse metadata of one graph edge record.
        Parameters
        ----------
        u: str
            Subject node curie identifier of the edge.
        v: str
            Subject node curie identifier of the edge.
        k: str
            Key identifier of the edge record (not used here).
        data: Dict
            Complete data dictionary of edge record fields.
        """
        # we blissfully assume that all the nodes of a
        # graph stream were analysed first by the MetaKnowledgeGraph
        # before the edges are analysed, thus we can test for
        # node 'n' existence internally, by identifier.
        #
        # Given the use case of multiple categories being assigned to a given node in a KGX data file,
        # either by category inheritance (ancestry all the way back up to NamedThing)
        # or by conflation (i.e. gene == protein id?), then the Cartesian product of
        # subject/object edges mappings need to be captured here.
        #
        self.edge_record_count += 1

        predicate: str = self._capture_predicate(data)
        if not predicate:
            # relationship needs a predicate to process?
            return

        if u not in self.node_catalog:
            _parse_warning("Edge 'subject' node ID", u, "not found in node catalog")
            # removing from edge count
            self.edge_record_count -= 1
            self.predicates[predicate] -= 1
            return

        for subj_cat_idx in self.node_catalog[u]:

            subject_category: str = self.Category.get_category_curie_from_index(
                subj_cat_idx
            )

            if v not in self.node_catalog:
                _parse_warning("Edge 'object' node ID", v, "not found in node catalog")
                self.edge_record_count -= 1
                self.predicates[predicate] -= 1
                return

            for obj_cat_idx in self.node_catalog[v]:

                object_category: str = self.Category.get_category_curie_from_index(
                    obj_cat_idx
                )

                self._process_triple(subject_category, predicate, object_category, data)

    def get_number_of_categories(self) -> int:
        """
        Counts the number of distinct (Biolink) categories encountered
        in the knowledge graph (not including those of 'unknown' category)

        Returns
        -------
        int
            Number of distinct (Biolink) categories found in the graph (excluding nodes with 'unknown' category)
        """
        # 'unknown' not tracked anymore...
        # return len([c for c in self.node_stats.keys() if c != 'unknown'])
        return len(self.node_stats.keys())

    def get_node_stats(self) -> Dict[str, Category]:
        """
        Returns
        -------
        Dict[str, Category]
            Statistics for the nodes in the graph.
        """
        # We no longer track 'unknown' node category counts - non TRAPI 1.1. compliant output
        # if 'unknown' in self.node_stats and not self.node_stats['unknown'].get_count():
        #     self.node_stats.pop('unknown')
        return self.node_stats

    def get_edge_stats(self) -> List[Dict[str, Any]]:
        """
        Returns
        -------
        List[Dict[str, Any]]
            Knowledge map for the list of edges in the graph.
        """
        # Not sure if this is "safe" but assume
        # that edge_stats may be cached once computed?
        if not self.edge_stats:
            for k, v in self.association_map.items():
                kedge = v
                relations = list(v["relations"])
                kedge["relations"] = relations
                self.edge_stats.append(kedge)
        return self.edge_stats

    def get_total_nodes_count(self) -> int:
        """
        Counts the total number of distinct nodes in the knowledge graph
        (**not** including those ignored due to being of 'unknown' category)

        Returns
        -------
        int
            Number of distinct nodes in the knowledge.
        """
        return len(self.node_catalog)

    def get_node_count_by_category(self, category_curie: str) -> int:
        """
        Counts the number of edges in the graph
        with the specified (Biolink) category curie.

        Parameters
        ----------
        category_curie: str
            Curie identifier for the (Biolink) category.

        Returns
        -------
        int
            Number of nodes for the given category.

        Raises
        ------
        RuntimeError
            Error if category identifier is empty string or None.
        """
        if not category_curie:
            raise RuntimeError(
                "get_node_count_by_category(): null or empty category argument!?"
            )
        if category_curie in self.node_stats.keys():
            return self.node_stats[category_curie].get_count()
        else:
            return 0

    def get_total_node_counts_across_categories(self) -> int:
        """
        The aggregate count of all node to category mappings for every category.
        Note that nodes with multiple categories will have their count replicated
        under each of its categories.

        Returns
        -------
        int
            Total count of node to category mappings for the graph.
        """
        count = 0
        for category in self.node_stats.values():
            count += category.get_count()
        return count

    def get_total_edges_count(self) -> int:
        """
        Gets the total number of 'valid' edges in the data set
        (ignoring those with 'unknown' subject or predicate category mappings)

        Returns
        ----------
        int
            Total count of edges in the graph.
        """
        return self.edge_record_count

    def get_edge_mapping_count(self) -> int:
        """
        Counts the number of distinct edge
        Subject (category) - P (predicate) -> Object (category)
        mappings in the knowledge graph.

        Returns
        ----------
        int
            Count of subject(category) - predicate -> object(category) mappings in the graph.
        """
        return len(self.get_edge_stats())

    def get_predicate_count(self) -> int:
        """
        Counts the number of distinct edge predicates
        in the knowledge graph.

        Returns
        ----------
        int
            Number of distinct (Biolink) predicates in the graph.
        """
        return len(self.predicates)

    def get_edge_count_by_predicate(self, predicate_curie: str) -> int:
        """
        Counts the number of edges in the graph with the specified predicate.

        Parameters
        ----------
        predicate_curie: str
            (Biolink) curie identifier for the predicate.

        Returns
        -------
        int
            Number of edges for the given predicate.

        Raises
        ------
        RuntimeError
            Error if predicate identifier is empty string or None.
        """
        if not predicate_curie:
            raise RuntimeError(
                "get_node_count_by_category(): null or empty predicate argument!?"
            )
        if predicate_curie in self.predicates:
            return self.predicates[predicate_curie]
        return 0

    def get_total_edge_counts_across_mappings(self) -> int:
        """
        Aggregate count of the edges in the graph for every mapping. Edges
        with subject and object nodes with multiple assigned categories will
        have their count replicated under each distinct mapping of its categories.

        Returns
        -------
        int
            Number of the edges counted across all mappings.
        """
        count = 0
        for edge in self.get_edge_stats():
            count += edge["count"]
        return count

    def get_edge_count_by_source(
        self,
        subject_category: str,
        predicate: str,
        object_category: str,
        facet: str = "knowledge_source",
        source: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Returns count by source for one S-P-O triple (S, O being Biolink categories; P, a Biolink predicate)
        """
        if not (subject_category and predicate and object_category):
            print(
                "Warning: get_edge_count_by_source() has some empty S-P-O arguments? Don't know what you are counting!",
                file=MetaKnowledgeGraph.error_log,
            )
            return dict()
        triple = (subject_category, predicate, object_category)
        if (
            triple in self.association_map
            and "count_by_source" in self.association_map[triple]
        ):
            if facet in self.association_map[triple]["count_by_source"]:
                if source:
                    if source in self.association_map[triple]["count_by_source"][facet]:
                        return self.association_map[triple]["count_by_source"][facet][
                            source
                        ]
                    else:
                        return dict()
                else:
                    return self.association_map[triple]["count_by_source"][facet]
            else:
                return dict()
        else:
            print(
                "Warning: get_edge_count_by_source(): unknown triple "
                + "("
                + subject_category
                + ","
                + predicate
                + ","
                + object_category
                + ")?",
                file=MetaKnowledgeGraph.error_log,
            )
            return dict()

    def summarize_graph_nodes(self, graph: BaseGraph) -> Dict:
        """
        Summarize the nodes in a graph.

        Parameters
        ----------
        graph: kgx.graph.base_graph.BaseGraph
            The graph

        Returns
        -------
        Dict
            The node stats
        """
        for n, data in graph.nodes(data=True):
            self.analyse_node(n, data)
        return self.get_node_stats()

    def summarize_graph_edges(self, graph: BaseGraph) -> List[Dict]:
        """
        Summarize the edges in a graph.

        Parameters
        ----------
        graph: kgx.graph.base_graph.BaseGraph
            The graph

        Returns
        -------
        List[Dict]
            The edge stats

        """
        for u, v, k, data in graph.edges(keys=True, data=True):
            self.analyse_edge(u, v, k, data)
        return self.get_edge_stats()

    def summarize_graph(self, graph: BaseGraph, name: str = None, **kwargs) -> Dict:
        """
        Generate a meta knowledge graph that describes the composition of the graph.

        Parameters
        ----------
        graph: kgx.graph.base_graph.BaseGraph
            The graph
        name: Optional[str]
            Name for the graph
        kwargs: Dict
            Any additional arguments (ignored in this method at present)

        Returns
        -------
        Dict
            A TRAPI 1.1 compliant meta knowledge graph of the knowledge graph returned as a dictionary.
        """
        if not self.graph_stats:
            node_stats = self.summarize_graph_nodes(graph)
            edge_stats = self.summarize_graph_edges(graph)
            # JSON sent back as TRAPI 1.1 version,
            # without the global 'knowledge_map' object tag
            self.graph_stats = {"nodes": node_stats, "edges": edge_stats}
            if name:
                self.graph_stats["name"] = name
            else:
                self.graph_stats["name"] = self.name
        return self.graph_stats

    def get_graph_summary(self, name: str = None, **kwargs) -> Dict:
        """
        Similar to summarize_graph except that the node and edge statistics are already captured
        in the MetaKnowledgeGraph class instance (perhaps by Transformer.process() stream inspection)
        and therefore, the data structure simply needs to be 'finalized' for saving or similar use.

        Parameters
        ----------
        name: Optional[str]
            Name for the graph (if being renamed)
        kwargs: Dict
            Any additional arguments (ignored in this method at present)

        Returns
        -------
        Dict
            A TRAPI 1.1 compliant meta knowledge graph of the knowledge graph returned as a dictionary.
        """
        if not self.graph_stats:
            # JSON sent back as TRAPI 1.1 version,
            # without the global 'knowledge_map' object tag
            self.graph_stats = {
                "nodes": self.get_node_stats(),
                "edges": self.get_edge_stats(),
            }
            if name:
                self.graph_stats["name"] = name
            else:
                self.graph_stats["name"] = self.name
        return self.graph_stats

    def save(self, file, name: str = None, file_format: str = "json") -> None:
        """
        Save the current MetaKnowledgeGraph to a specified (open) file (device).

        Parameters
        ----------
        file: File
            Text file handler open for writing.
        name: str
            Optional string to which to (re-)name the graph.
        file_format:  str
            Text output format ('json' or 'yaml') for the saved meta knowledge graph  (default: 'json')

        Returns
        -------
        None
        """
        stats = self.get_graph_summary(name)
        if not file_format or file_format == "json":
            dump(stats, file, indent=4, default=mkg_default)
        else:
            yaml.dump(stats, file)


def generate_meta_knowledge_graph(graph: BaseGraph, name: str, filename: str) -> None:
    """
    Generate a knowledge map that describes the composition of the graph
    and write to ``filename``.

    Parameters
    ----------
    graph: kgx.graph.base_graph.BaseGraph
        The graph
    name: Optional[str]
        Name for the graph
    filename: str
        The file to write the knowledge map to

    """
    graph_stats = summarize_graph(graph, name)
    with open(filename, mode="w") as mkgh:
        dump(graph_stats, mkgh, indent=4, default=mkg_default)


def summarize_graph(graph: BaseGraph, name: str = None, **kwargs) -> Dict:
    """
    Generate a meta knowledge graph that describes the composition of the graph.

    Parameters
    ----------
    graph: kgx.graph.base_graph.BaseGraph
        The graph
    name: Optional[str]
        Name for the graph
    kwargs: Dict
        Any additional arguments

    Returns
    -------
    Dict
        A TRAPI 1.1 compliant meta knowledge graph of the knowledge graph returned as a dictionary.
    """
    mkg = MetaKnowledgeGraph(name)
    return mkg.summarize_graph(graph)
